try:
    import h5py
except Exception:
    pass

import os
import logging
import numpy as np
from pysnptools.pstreader import PstReader
from pysnptools.pstreader.pstdata import PstData
import warnings


class PstHdf5(PstReader):
    """
    A :class:`.PstReader` for reading \*.pst.hdf5 files from disk.

    See :class:`.PstReader` for general examples of using PstReaders.

    The general HDF5 format is described in `here <http://www.hdfgroup.org/HDF5/>`__. The PstHdf5 format stores
    val, row, col, row_property, and col_property information in Hdf5 format.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The PstHdf5 file to read.

        :Example:

        >>> from pysnptools.pstreader import PstHdf5
        >>> from pysnptools.util import example_file # Download and return local file name
        >>>
        >>> psthdf5_file = example_file('pysnptools/examples/toydata.iidmajor.snp.hdf5')
        >>> on_disk = PstHdf5(psthdf5_file) # PstHdf5 can load .pst.hdf5, .snp.hdf5, and kernel.hdf5
        >>> print(on_disk.row_count)
        500

    **Methods beyond** :class:`.PstReader`
    """

    def __init__(self, filename, block_size=5000):  # LATER document block size
        super(PstHdf5, self).__init__()  # We know PstReader doesn't want the file name

        self._block_size = block_size

        self._ran_once = False
        self._h5 = None

        self.filename = str(filename)

    def __getstate__(self):
        return self.filename, self._block_size

    def __setstate__(self, state):
        filename, block_size = state
        self.__init__(filename, block_size=block_size)

    def __repr__(self):
        return "{0}('{1}')".format(self.__class__.__name__, self.filename)
        # !!LATER print non-default values, too

    def copyinputs(self, copier):
        copier.input(self.filename)

    @property
    def row(self):
        self._run_once()
        return self._row

    @property
    def col(self):
        self._run_once()
        return self._col

    @property
    def row_property(self):
        self._run_once()
        return self._row_property

    @property
    def col_property(self):
        self._run_once()
        return self._col_property

    def _find_vocab(self):
        vocab_list = [
            ["row", "col", "val", "row_property", "col_property"],
            ["iid", "sid", "val", None, "pos"],
            ["iid", "rs", "snps", None, "pos"],
        ]

        for vocab in vocab_list:
            if all((key is None or key in self._h5) for key in vocab):
                return vocab
        raise Exception(
            "Don't know how to read HDF5 with these keys: "
            + ",".join(self._h5.iterkeys())
        )

    def _run_once(self):
        if self._ran_once:
            return
        try:
            self._h5 = h5py.File(self.filename, "r")
        except IOError as e:
            raise IOError(
                "Missing or unopenable file '{0}' -- Native error message: {1}".format(
                    self.filename, e
                )
            )

        (
            row_key,
            col_key,
            val_key,
            row_property_key,
            col_property_key,
        ) = self._find_vocab()

        self._row = PstData._fixup_input(self._h5[row_key])
        self._col = PstData._fixup_input(self._h5[col_key])
        if self._row.dtype == self._col.dtype and np.array_equal(
            self._row, self._col
        ):  # If it's square, mark it so by making the col and row the same object
            self._row = self._col
        self._row_property = PstData._fixup_input(
            self._h5[row_property_key] if row_property_key else None,
            count=len(self._row),
        )  # Extra "if ... else" for backwards compatibility.
        self._col_property = PstData._fixup_input(
            self._h5[col_property_key], count=len(self._col)
        )
        self.val_in_file = self._h5[val_key]

        self.is_col_major = None
        if "col-major" in self.val_in_file.attrs:
            self.is_col_major = self.val_in_file.attrs["col-major"]
        elif "SNP-major" in self.val_in_file.attrs:
            self.is_col_major = self.val_in_file.attrs["SNP-major"]
        assert (
            self.is_col_major is not None
        ), "In Hdf5 the 'val' matrix must have a Boolean 'col-major' (or 'SNP-major') attribute"

        S_original = len(self._col)
        N_original = len(self._row)
        if self.is_col_major:
            if not self.val_in_file.shape[-2:] == (S_original, N_original):
                raise Exception(
                    "In Hdf5, the val matrix dimensions don't match those of 'row' and 'col'"
                )
            self._val_shape = (
                None if len(self.val_in_file.shape) == 2 else self.val_in_file.shape[0]
            )
        else:
            if not self.val_in_file.shape[:2] == (N_original, S_original):
                raise Exception(
                    "In Hdf5, the val matrix dimensions don't match those of 'row' and 'col'"
                )
            self._val_shape = (
                None if len(self.val_in_file.shape) == 2 else self.val_in_file.shape[-1]
            )
        self._ran_once = True

    @staticmethod
    def _is_sorted_without_repeats(list):
        if len(list) < 2:
            return True
        for i in range(1, len(list)):
            if not list[i - 1] < list[i]:
                return False
        return True

    def flush(self):
        """Flush the connection to the HDF5 file. (If values or properties are accessed again, the file will be reopened.)

        >>> import pysnptools.util as pstutil
        >>> from pysnptools.pstreader import PstHdf5
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> hdf5_file = example_file("pysnptools/examples/toydata.kernel.hdf5")
        >>> reader = PstHdf5(hdf5_file)
        >>> val1 = reader[0,0].read()
        >>> reader.flush()

        """
        if self._ran_once:
            self._h5.close()
            del self._h5
            self._h5 = None
            self._ran_once = False

    def __del__(self):
        if (
            self._h5 is not None
        ):  # we need to test this because Python doesn't guarantee that __init__ was fully run
            self._h5.close()

    def _read_direct(self, val, val_order, selection=None):
        if selection is None:
            selection = np.s_[:, :] if self._val_shape is None else np.s_[:, :, :]

        if self.is_col_major:
            selection = tuple(reversed(selection))

        if val_order == "F":
            self.val_in_file.read_direct(val.T, selection)
        else:
            assert val_order == "C", "real assert"
            self.val_in_file.read_direct(val, selection)

    def _create_block(self, block_size, order, dtype):
        matches_order = self.is_col_major == (order == "F")
        opposite_order = "C" if order == "F" else "F"
        dtype = np.dtype(dtype)

        if matches_order:
            if self._val_shape is None:
                return (
                    np.empty([len(self._row), block_size], dtype=dtype, order=order),
                    order,
                )
            else:
                return (
                    np.empty(
                        [len(self._row), block_size, self._val_shape],
                        dtype=dtype,
                        order=order,
                    ),
                    order,
                )
        else:
            if self._val_shape is None:
                # LATER should make own version np.empty with optional 3rd dimension and np.nan fill?
                return (
                    np.empty(
                        [len(self._row), block_size], dtype=dtype, order=opposite_order
                    ),
                    opposite_order,
                )
            else:
                return (
                    np.empty(
                        [
                            len(self._row),
                            block_size,
                            self._val_shape,
                        ],
                        dtype=dtype,
                        order=opposite_order,
                    ),
                    opposite_order,
                )

    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        self._run_once()
        dtype = np.dtype(dtype)

        assert order in ["F", "C", "A"], "Expect order to be 'F', 'C' or 'A'"

        if order == "A":
            order = "F" if self.is_col_major else "C"

        opposite_order = "C" if order == "F" else "F"

        if row_index_or_none is not None:
            row_index_count = len(row_index_or_none)
            row_index_list = row_index_or_none
            row_is_sorted = PstHdf5._is_sorted_without_repeats(row_index_list)
            if isinstance(row_index_list, np.ndarray):
                row_index_list = row_index_list.tolist()
        else:
            row_index_count = self.row_count
            row_index_list = range(self.row_count)
            row_is_sorted = True

        if col_index_or_none is not None:
            col_index_count = len(col_index_or_none)
            col_index_list = col_index_or_none
            if isinstance(col_index_list, np.ndarray):
                col_index_list = col_index_list.tolist()
        else:
            col_index_count = self.col_count
            col_index_list = list(range(self.col_count))
        # Check if snps and iids indexes are in order and in range
        col_are_sorted = PstHdf5._is_sorted_without_repeats(col_index_list)

        if self._val_shape is None:
            val = np.empty([row_index_count, col_index_count], dtype=dtype, order=order)
        else:
            val = np.empty(
                [row_index_count, col_index_count, self._val_shape],
                dtype=dtype,
                order=order,
            )
        val.fill(np.nan)  # LATER Keep this?

        matches_order = self.is_col_major == (order == "F")
        is_simple = (
            not force_python_only and row_is_sorted and col_are_sorted and matches_order
        )  # If 'is_simple' may be able to use a faster reader

        # case 0 -- zero elements in val
        if row_index_count == 0 or col_index_count == 0:
            pass

        # case 1 - all cols & all rows requested
        elif (
            is_simple
            and col_index_count == self.col_count
            and row_index_count == self.row_count
        ):
            self._read_direct(val, order)

        # case 2 - some cols and all rows
        elif is_simple and row_index_count == self.row_count:
            selection = (
                np.s_[:, col_index_list]
                if self._val_shape is None
                else np.s_[:, col_index_list, :]
            )
            self._read_direct(val, order, selection)

        # case 3 all cols and some row
        elif is_simple and col_index_count == self.col_count:
            selection = (
                np.s_[row_index_list, :]
                if self._val_shape is None
                else np.s_[row_index_list, :, :]
            )
            self._read_direct(val, order, selection)

        # case 4 some cols and some rows -- use blocks
        else:
            block_size = min(self._block_size, col_index_count)
            block, block_order = self._create_block(block_size, order, dtype)

            if not col_are_sorted:
                col_index_index_list = np.argsort(col_index_list).tolist()
                col_index_list_sorted = list(
                    np.array(col_index_list)[col_index_index_list]
                )
            else:
                col_index_index_list = np.arange(col_index_count)
                col_index_list_sorted = col_index_list

            for start in range(0, col_index_count, block_size):
                # print start
                stop = min(start + block_size, col_index_count)
                if (
                    stop - start < block_size
                ):  # On the last loop, the buffer might be too big, so make it smaller
                    block, block_order = self._create_block(stop - start, order, dtype)
                col_index_list_forblock = col_index_list_sorted[start:stop]
                col_index_index_list_forblock = col_index_index_list[start:stop]
                selection = (
                    np.s_[:, col_index_list_forblock]
                    if self._val_shape is None
                    else np.s_[:, col_index_list_forblock, :]
                )
                self._read_direct(block, block_order, selection)
                val[:, col_index_index_list_forblock] = block[row_index_list, :]

        # !!LATER does this test work when the size is 1 x 1 and order if F? iid_index_or_none=[0], sid_index_or_none=[1000] (based on test_blocking_hdf5)
        has_right_order = (order == "C" and val.flags["C_CONTIGUOUS"]) or (
            order == "F" and val.flags["F_CONTIGUOUS"]
        )
        expected_shape = (
            (row_index_count, col_index_count)
            if self._val_shape is None
            else (row_index_count, col_index_count, self._val_shape)
        )
        assert val.shape == expected_shape and val.dtype == dtype and has_right_order
        return val

    @staticmethod
    def write(filename, pstdata, hdf5_dtype=None, col_major=True):
        """Writes a :class:`PstData` to PstHdf5 format and returns the :class:`.PstHdf5`.

        :param filename: the name of the file to create
        :type filename: string
        :param pstdata: The in-memory data that should be written to disk.
        :type pstdata: :class:`PstData`
        :param hdf5_dtype: None (use the .val's dtype) or a Hdf5 dtype, e.g. 'f8','f4',etc.
        :type hdf5_dtype: string
        :param col_major: Tells if vals should be stored on disk in col_major (default) or row_major format.
        :type col_major: bool
        :rtype: :class:`.PstHdf5`

        >>> import numpy as np
        >>> from pysnptools.pstreader import PstData, PstHdf5
        >>> import pysnptools.util as pstutil
        >>> data1 = PstData(row=['a','b','c'],col=['y','z'],val=[[1,2],[3,4],[np.nan,6]],row_property=['A','B','C'])
        >>> pstutil.create_directory_if_necessary("tempdir/tiny.pst.hdf5")
        >>> PstHdf5.write("tempdir/tiny.pst.hdf5",data1)          # Write data in PstHdf5 format
        PstHdf5('tempdir/tiny.pst.hdf5')
        """

        assert hdf5_dtype is None or (
            isinstance(hdf5_dtype, str)
            and len(hdf5_dtype) == 2
            and hdf5_dtype[0] == "f"
        ), "Expect hdf5_dtype to be None or to start with 'f', e.g. 'f4' for single, 'f8' for double"

        val = (pstdata.val.T) if col_major else pstdata.val

        def any_u_to_a(possible_unicode):
            # If it's any kind of string, encode it as ascii
            if (
                possible_unicode.dtype.char == "S" or possible_unicode.dtype.char == "U"
            ):  # not using np.issubdtype because of future warning
                return np.array(possible_unicode, dtype="S")
            else:  # Otherwise, just leave it.
                return possible_unicode

        with h5py.File(filename, "w") as h5:
            h5.create_dataset("row", data=any_u_to_a(pstdata.row))
            h5.create_dataset("col", data=any_u_to_a(pstdata.col))
            h5.create_dataset("row_property", data=any_u_to_a(pstdata.row_property))
            h5.create_dataset("col_property", data=any_u_to_a(pstdata.col_property))
            h5.create_dataset(
                "val", data=val, dtype=hdf5_dtype, shuffle=True
            )  # compression="gzip", doesn't seem to work with Anaconda
            h5["val"].attrs["col-major"] = col_major

        return PstHdf5(filename)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
from pysnptools.pstreader import PstReader


class _PstSubset(PstReader):
    def __init__(self, internal, row_indexer, col_indexer):
        """
        an indexer can be:
             an integer i (same as [i])
             a slice
             a list of integers (including negatives)
             a list of Booleans
        """
        super(_PstSubset, self).__init__()
        self._ran_once = False

        self._internal = internal
        self._row_indexer = PstReader._make_sparray_or_slice(row_indexer)
        self._col_indexer = PstReader._make_sparray_or_slice(col_indexer)

    def __repr__(self):
        s = "{0}[{1},{2}]".format(
            self._internal,
            _PstSubset.static_nice_string(self, self._row_indexer),
            _PstSubset.static_nice_string(self, self._col_indexer),
        )
        return s

    def copyinputs(self, copier):
        self._internal.copyinputs(copier)

    @property
    def row(self):
        self._run_once()
        return self._row

    @property
    def col(self):
        self._run_once()
        return self._col

    @property
    def row_property(self):
        self._run_once()
        return self._row_property

    @property
    def col_property(self):
        self._run_once()
        return self._col_property

    # Most _read's support only indexlists or None, but this one supports Slices, too.
    _read_accepts_slices = True

    def _read(
        self,
        row_indexer,
        col_indexer,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        self._run_once()
        dtype = np.dtype(dtype)
        if hasattr(self._internal, "_read_accepts_slices"):
            assert (
                self._internal._read_accepts_slices
            ), "If an object has the _read_accepts_slices attribute, it must have value 'True'"
            composed_row_index_or_none = _PstSubset.compose_indexer_with_indexer(
                self._internal.row_count, self._row_indexer, self.row_count, row_indexer
            )
            composed_col_index_or_none = _PstSubset.compose_indexer_with_indexer(
                self._internal.col_count, self._col_indexer, self.col_count, col_indexer
            )
            val = self._internal._read(
                composed_row_index_or_none,
                composed_col_index_or_none,
                order,
                dtype,
                force_python_only,
                view_ok,
                num_threads,
            )
            return val
        else:
            row_index_or_none = PstReader._make_sparray_from_sparray_or_slice(
                self.row_count, row_indexer
            )
            composed_row_index_or_none = _PstSubset.compose_indexer_with_index_or_none(
                self._internal.row_count,
                self._row_indexer,
                self.row_count,
                row_index_or_none,
            )
            col_index_or_none = PstReader._make_sparray_from_sparray_or_slice(
                self.col_count, col_indexer
            )
            composed_col_index_or_none = _PstSubset.compose_indexer_with_index_or_none(
                self._internal.col_count,
                self._col_indexer,
                self.col_count,
                col_index_or_none,
            )
            val = self._internal._read(
                composed_row_index_or_none,
                composed_col_index_or_none,
                order,
                dtype,
                force_python_only,
                view_ok,
                num_threads,
            )
            return val

    def _run_once(self):
        if self._ran_once:
            return

        self._ran_once = True
        self._row = self._internal.row[
            self._row_indexer
        ]  # !!! would be nice if these four calls were replaced by calls to, e.g. "def _col_index(self,col_indexer)", etc that could be overridden (by for example Pairs)
        self._col = self._internal.col[self._col_indexer]
        if self._row.dtype == self._col.dtype and np.array_equal(
            self._row, self._col
        ):  # When an object is square, keep the row and col the same object.
            self._col = self._row
        self._row_property = self._internal.row_property[self._row_indexer]
        self._col_property = self._internal.col_property[self._col_indexer]

    _slice_format = {
        (False, False, False): ":",
        (False, False, True): "::{2}",
        (False, True, False): ":{1}",
        (False, True, True): ":{1}:{2}",
        (True, False, False): "{0}:",
        (True, False, True): "{0}::{2}",
        (True, True, False): "{0}:{1}",
        (True, True, True): "{0}:{1}:{2}",
    }

    @staticmethod
    def static_nice_string(self, some_slice):
        if isinstance(some_slice, slice):
            return _PstSubset._slice_format[
                (
                    some_slice.start is not None,
                    some_slice.stop is not None,
                    some_slice.step is not None,
                )
            ].format(some_slice.start, some_slice.stop, some_slice.step)
        elif len(some_slice) == 1:
            return str(some_slice[0])
        elif len(some_slice) < 10:
            return "[{0}]".format(",".join([str(i) for i in some_slice]))
        else:
            return "[{0},...]".format(",".join([str(i) for i in some_slice[:10]]))

    # !!commented out because doesn't guarantee that the shortcut will return with the dtype and order requested.
    #                  Also, didn't handle stacked do-nothing subsets
    # def read(self, order='F', dtype=np.float64, force_python_only=False, view_ok=False):
    #    if view_ok and hasattr(self._internal,"val") and _PstSubset._is_all_slice(self._row_indexer) and _PstSubset._is_all_slice(self._col_indexer):
    #        return self._internal
    #    else:
    #        return PstReader.read(self, order, dtype, force_python_only, view_ok)

    @staticmethod
    def compose_indexer_with_index_or_none(countA, indexerA, countB, index_or_noneB):
        if _PstSubset._is_all_slice(indexerA):
            return index_or_noneB

        indexA = PstReader._make_sparray_from_sparray_or_slice(countA, indexerA)

        if _PstSubset._is_all_slice(index_or_noneB):
            return indexA

        indexAB = indexA[index_or_noneB]

        return indexAB

    @staticmethod
    def compose_indexer_with_indexer(countA, indexerA, countB, indexerB):
        if _PstSubset._is_all_slice(indexerA):
            return indexerB

        if _PstSubset._is_all_slice(indexerB):
            return indexerA

        indexA = PstReader._make_sparray_from_sparray_or_slice(countA, indexerA)
        indexB = PstReader._make_sparray_from_sparray_or_slice(countB, indexerB)

        indexAB = indexA[indexB]

        return indexAB



"""Tools for reading and manipulating SNP data.
"""

def _snps_fixup(snp_input, iid_if_none=None,count_A1=None):
    if isinstance(snp_input, str):
        return Bed(snp_input,count_A1=count_A1)

    if isinstance(snp_input, dict):
        return SnpData(iid=snp_input['iid'],sid=snp_input['header'],val=snp_input['vals'])

    if snp_input is None:
        assert iid_if_none is not None, "snp_input cannot be None here"
        return SnpData(iid_if_none, sid=np.empty((0),dtype='str'), val=np.empty((len(iid_if_none),0)),pos=np.empty((0,3)),name="") #todo: make a static factory method on SnpData

    return snp_input


from pysnptools.snpreader.snpreader import SnpReader
from pysnptools.snpreader.snpdata import SnpData
from pysnptools.snpreader.bed import Bed, plink_chrom_map, reverse_plink_chrom_map
from pysnptools.snpreader.ped import Ped
from pysnptools.snpreader.dat import Dat
from pysnptools.snpreader.snphdf5 import SnpHdf5
from pysnptools.snpreader.snphdf5 import Hdf5
from pysnptools.snpreader.snpnpz import SnpNpz
from pysnptools.snpreader.dense import Dense
from pysnptools.snpreader.pheno import Pheno
from pysnptools.snpreader.snpmemmap import SnpMemMap
from pysnptools.snpreader._mergesids import _MergeSIDs
from pysnptools.snpreader._mergeiids import _MergeIIDs
from pysnptools.snpreader.snpgen import SnpGen
from pysnptools.snpreader.distributedbed import DistributedBed, _Distributed1Bed



from pysnptools.util.mapreduce1.runner import *
import logging
import unittest
import io
import sys

class DistributableTest(object) : #implements IDistributable
    '''
    This is a class for distributing any testing.
    '''
    def __init__(self,test_or_suite,tempdirectory=None):
        self._test_or_suite = test_or_suite
        self.__tempdirectory = tempdirectory
        self.name = str(test_or_suite)

    @staticmethod
    def deep_iter(test_or_suite):
        try:
            for sub in test_or_suite:
                for subsub in DistributableTest.deep_iter(sub):
                    yield subsub
        except TypeError as detail:
            yield test_or_suite

    def do_work(self, test):

        ###############test_result = unittest.TestResult()
        r = unittest.TextTestRunner(failfast=False)
        logging.info(test)
        test.setUpClass()
        test_result = r.run(test)
        ##############test.run(result=test_result)
        test.tearDownClass()
        return test_result

 #start of IDistributable interface--------------------------------------

    _work_count = None
    @property
    def work_count(self):
        if self._work_count is None:
            self._work_count = 0
            for test in self.deep_iter(self._test_or_suite):
                self._work_count += 1
        return self._work_count


    def work_sequence(self):
        for i, test in enumerate(self.deep_iter(self._test_or_suite)):
            yield lambda test=test : self.do_work(test)  # the 'test=test' is need to get around a strangeness in Python

    def reduce(self, result_sequence):
        fp = io.StringIO() if sys.version_info >= (3,0) else io.BytesIO()
        error_count = 0
        failure_count = 0
        test_result_list = []
        for test_result in result_sequence:
            test_result_list.append(test_result)
            for error in test_result.errors:
                error_count += 1
                fp.write("Error # {0} - {1}\n{2}\n============\n".format(error_count, error[0],error[1]))
            for failure in test_result.failures:
                failure_count += 1
                fp.write("Error # {0} - {1}\n{2}\n============\n".format(failure_count, failure[0],failure[1]))

        fp.write("\nFinal Error # {0}. Final Failure # {1}\n".format(error_count,failure_count))
        s = fp.getvalue()
        fp.close()

        print(s)
        return s

    @property
    def tempdirectory(self):
        return self.__tempdirectory

    #optional override -- the str name of the instance is used by the cluster as the job name
    def __str__(self):
        return "{0}(...)".format(self.__class__.__name__)
 #end of IDistributable interface---------------------------------------




import numpy as np
import logging
import doctest
import shutil
import cloudpickle as pickle

from pysnptools.snpreader import Bed
from pysnptools.snpreader import SnpHdf5, SnpNpz
from pysnptools.snpreader import Dat
from pysnptools.snpreader import Dense
from pysnptools.snpreader import Pheno
from pysnptools.snpreader import Ped
from pysnptools.snpreader import DistributedBed
from pysnptools.standardizer import Unit
from pysnptools.standardizer import Beta
from pysnptools.util import create_directory_if_necessary
from pysnptools.kernelreader.test import TestKernelReader
from pysnptools.kernelreader.test import TestKrDocStrings
from pysnptools.pstreader.test import TestPstReader
from pysnptools.pstreader.test import TestPstDocStrings
from pysnptools.pstreader.pstmemmap import TestPstMemMap
from pysnptools.snpreader.snpmemmap import TestSnpMemMap
from pysnptools.snpreader.snpgen import TestSnpGen
from pysnptools.snpreader.distributedbed import TestDistributedBed
from pysnptools.util.generate import TestGenerate
from pysnptools.util._example_file import TestExampleFile
from pysnptools.kernelreader.test import _fortesting_JustCheckExists
from pysnptools.util.intrangeset import TestIntRangeSet
from pysnptools.util.test import TestUtilTools
from pysnptools.util.filecache.test import TestFileCache
from pysnptools.distreader.test import TestDistReaders

import unittest
import os.path
import time


class TestPySnpTools(unittest.TestCase):
    def xtest_aaa_hdf5_speed(self):  # !!too slow to use all the time
        # currentFolder + "/examples/toydata.5chrom.bed"
        # currentFolder + "/examples/delme.hdf5"
        bedFileName = r"c:\Source\carlk\fastlmm\tests\datasets\all_chr.maf0.001.N300"  # !! local paths
        hdf5Pattern = r"c:\Source\carlk\fastlmm\tests\datasets\del.{0}.hdf5"  # !!
        tt0 = time.time()
        snpreader_bed = Bed(bedFileName, count_A1=False)

        S0 = snpreader_bed.sid_count
        snp_index_list0 = range(min(S0, 15000))

        hdf5FileName = hdf5Pattern.format(len(snp_index_list0))

        snpDataBed = snpreader_bed[:, snp_index_list0].read()
        tt1 = time.time()
        logging.info("Read bed %.2f seconds" % (tt1 - tt0))
        SnpHdf5.write(hdf5FileName, snpDataBed)
        tt2 = time.time()
        logging.info("write SnpHdf5 bed %.2f seconds" % (tt2 - tt1))

        snpreader_hdf5 = SnpHdf5(hdf5FileName)
        assert snpreader_hdf5.iid_count == snpreader_bed.iid_count
        S = snpreader_hdf5.sid_count
        N_original = snpreader_hdf5.iid_count
        iid_index_list = sorted(range(N_original - 1, 0, -2))

        snp_index_list = sorted(range(S - 1, 0, -2))  # !!
        # snp_index_list = range(S//2)

        snpreader_hdf5 = snpreader_hdf5[iid_index_list, :]
        _ = snpreader_hdf5[:, snp_index_list].read()
        tt3 = time.time()
        logging.info(
            "read SnpHdf5 with reversed indexes bed %.2f seconds" % (tt3 - tt2)
        )

        _ = snpreader_hdf5[:, snp_index_list].read(order="C")
        tt4 = time.time()
        logging.info(
            "read SnpHdf5 C with reversed indexes bed %.2f seconds" % (tt4 - tt3)
        )

        print("the end")

    @classmethod
    def setUpClass(self):
        self.currentFolder = os.path.dirname(os.path.realpath(__file__))
        # TODO: get data set with NANs!
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        assert snpreader.pos[0, 1] != snpreader.pos[0, 1]  # real assert
        self.pheno_fn = self.currentFolder + "/examples/toydata.phe"
        self.snpdata = snpreader.read(order="F", force_python_only=True)
        self.snps = self.snpdata.val

    def test_write_bad_value_and_good(self):
        from pysnptools.snpreader import Pheno, Bed, SnpData
        import pysnptools.util as pstutil
        from pysnptools.util import example_file  # Download and return local file name

        pheno_fn = example_file("pysnptools/examples/toydata.phe")
        snpdata = Pheno(pheno_fn).read()  # Read data from Pheno format
        pstutil.create_directory_if_necessary("tempdir/toydata.5chrom.bed")
        error_seen = False
        try:
            Bed.write(
                "tempdir/toydata.5chrom.bed", snpdata, count_A1=False
            )  # Write data in Bed format
        except Exception:
            error_seen = True
        assert error_seen
        # Can write from an int8 array, too.
        snpdata_int = SnpData(
            val=np.int_(snpdata.val).astype("int8"),
            iid=snpdata.iid,
            sid=snpdata.sid,
            pos=snpdata.pos,
            _require_float32_64=False,
        )
        assert snpdata_int.val.dtype == "int8"
        error_seen = False
        try:
            Bed.write(
                "tempdir/toydata.5chrom.bed",
                snpdata_int,
                count_A1=False,
                _require_float32_64=False,
            )
        except Exception:
            error_seen = True
        assert error_seen

        bed_fn = example_file("pysnptools/examples/toydata.5chrom.*", "*.bed")
        snpdata = Bed(bed_fn)[:, ::2].read()  # Read every-other SNP
        pstutil.create_directory_if_necessary("tempdir/everyother.bed")
        Bed.write(
            "tempdir/everyother.bed", snpdata, count_A1=False
        )  # Write data in Bed format
        # Bed('tempdir/everyother.bed',count_A1=False)
        # Can write from an int8 array, too.
        snpdata_int = SnpData(
            val=np.int_(snpdata.val).astype("int8"),
            iid=snpdata.iid,
            sid=snpdata.sid,
            pos=snpdata.pos,
            _require_float32_64=False,
        )
        snpdata_int.val.dtype
        assert snpdata_int.val.dtype == "int8"
        Bed.write(
            "tempdir/everyother.bed",
            snpdata_int,
            count_A1=False,
            _require_float32_64=False,
        )
        # Bed('tempdir/everyother',count_A1=False)

    def test_val_assign(self):
        from pysnptools.snpreader import SnpData
        from pysnptools.kernelreader import KernelData

        iid_count = 3
        sid_count = 2
        iid = [[str(i), str(i)] for i in range(iid_count)]
        sid = [str(s) for s in range(sid_count)]
        np.random.seed(0)
        vali = np.random.randint(10, size=[iid_count, sid_count])
        vali2 = np.random.randint(10, size=[iid_count, sid_count])
        snpdata = SnpData(iid=iid, sid=sid, val=vali)
        snpdata.val = vali2
        assert snpdata.val.dtype == np.float64

        valk = np.random.randint(10, size=[iid_count, iid_count])
        valk2 = np.random.randint(10, size=[iid_count, iid_count])
        kerneldata = KernelData(iid=iid, val=valk)
        assert np.float64 == kerneldata.val.dtype
        kerneldata.val = valk2
        assert np.float64 == kerneldata.val.dtype

        valstr = np.array(vali, dtype=np.str_)
        valstr[0, 0] = "cannot convert"
        error_seen = False
        try:
            snpdata.val = valstr
        except Exception:
            error_seen = True
        assert error_seen
        error_seen = False
        try:
            kerneldata.val = vali
        except Exception:
            error_seen = True
        assert error_seen
        error_seen = False

    def test_diagKtoN(self):
        """
        make sure standardization on SNPs results in sum(diag(K))=N
        """

        np.random.seed(42)
        m = np.random.random((100, 1000))
        from pysnptools.standardizer import DiagKtoN

        s = DiagKtoN()
        s.standardize(m)
        K = m.dot(m.T)
        sum_diag = np.sum(np.diag(K))

        np.testing.assert_almost_equal(100, sum_diag)

    def test_c_reader_bed(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        self.c_reader(snpreader)

    def test_c_reader_bed_count_A1(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=True
        )
        snpdata = snpreader.read()
        snpdata.val = 2 - snpdata.val
        self.c_reader(snpdata)

    def test_p_reader_bed(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        ).read(force_python_only=True)
        self.c_reader(snpreader)

    def test_p_reader_bed_count_A1(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=True
        )
        snpdata = snpreader.read(force_python_only=True)
        snpdata.val = 2 - snpdata.val
        self.c_reader(snpdata)

    def test_bed_2021(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed",
            count_A1=True,
            fam_filename=self.currentFolder + "/examples/toydata.5chrom.2.fam",
            bim_filename=self.currentFolder + "/examples/toydata.5chrom.XYMT.bim",
        )
        snpdata = snpreader.read()

        from pysnptools.snpreader import reverse_plink_chrom_map
        import pysnptools.util as pstutil

        pstutil.create_directory_if_necessary("tempdir/test_reverse_chrom_map.bed")
        Bed.write(
            "tempdir/test_reverse_chrom_map.bed",
            snpdata,
            count_A1=False,
            reverse_chrom_map=reverse_plink_chrom_map,
        )

        see_exception = False
        try:
            snpreader = Bed(
                self.currentFolder + "/examples/toydata.5chrom.bed",
                count_A1=True,
                bim_filename=self.currentFolder + "/examples/toydata.5chrom.bad.bim",
            )
            snpdata = snpreader.read()
        except ValueError:
            see_exception = True
        assert see_exception

        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed",
            count_A1=True,
            bim_filename=self.currentFolder + "/examples/toydata.5chrom.bad.bim",
            chrom_map={"B": 1, "A": 2, "D": 3},
        )
        snpdata = snpreader.read()

    def test_bed_int8(self):
        snpreader = Bed(
            self.currentFolder + "/../tests/datasets/distributed_bed_test1_X.bed",
            count_A1=True,
        )
        ref = snpreader.read()
        ref.val[ref.val != ref.val] = -127
        ref.val = ref.val.astype("int8")
        for force_python_only in [False, True]:
            for order in ["F", "C"]:
                snpdata = snpreader.read(
                    dtype="int8",
                    _require_float32_64=False,
                    force_python_only=force_python_only,
                    order=order,
                )
                assert snpdata.val.dtype == "int8"
                assert np.allclose(snpdata.val, ref.val, equal_nan=True)
                output = "tempdir/snpreader/int8.bed"
                create_directory_if_necessary(output)
                for count_A1 in [False, True]:
                    bed2 = Bed.write(
                        output,
                        snpdata,
                        count_A1=count_A1,
                        _require_float32_64=False,
                        force_python_only=force_python_only,
                    )
                    assert np.allclose(
                        bed2.read(
                            dtype="int8",
                            _require_float32_64=False,
                            force_python_only=force_python_only,
                        ).val,
                        ref.val,
                        equal_nan=True,
                    )

    def test_scalar_index(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        arr = np.int64(1)
        snpreader[arr, arr]

    def test_c_reader_hdf5(self):
        snpreader = SnpHdf5(self.currentFolder + "/examples/toydata.snpmajor.snp.hdf5")
        self.c_reader(snpreader)

    def test_hdf5_case3(self):
        snpreader1 = SnpHdf5(
            self.currentFolder + "/examples/toydata.snpmajor.snp.hdf5"
        )[::2, :]
        snpreader2 = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )[::2, :]
        self.assertTrue(
            np.allclose(
                snpreader1.read().val, snpreader2.read().val, rtol=1e-05, atol=1e-05
            )
        )

    def test_standardize_hdf5(self):
        snpreader = SnpHdf5(self.currentFolder + "/examples/toydata.iidmajor.snp.hdf5")
        self.standardize(snpreader)

    def test_c_reader_npz(self):
        snpreader = SnpNpz(self.currentFolder + "/examples/toydata10.snp.npz")
        snpdata = snpreader.read(order="F", force_python_only=False)
        snp_c = snpdata.val

        self.assertEqual(np.float64, snp_c.dtype)
        self.assertTrue(np.allclose(self.snps[:, :10], snp_c, rtol=1e-05, atol=1e-05))

        snpreader1 = SnpNpz(self.currentFolder + "/examples/toydata10.snp.npz")
        snpreader2 = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )[:, :10]
        self.assertTrue(
            np.allclose(
                snpreader1.read().val, snpreader2.read().val, rtol=1e-05, atol=1e-05
            )
        )

        snpdata.val[
            1, 2
        ] = np.NaN  # Inject a missing value to test writing and reading missing values
        output = "tempdir/snpreader/toydata10.snp.npz"
        create_directory_if_necessary(output)
        SnpNpz.write(output, snpdata)
        snpdata2 = SnpNpz(output).read()
        np.testing.assert_array_almost_equal(snpdata.val, snpdata2.val, decimal=10)

        snpdata3 = snpdata[:, 0:0].read()  # create snpdata with no sids
        output = "tempdir/snpreader/toydata0.snp.npz"
        SnpNpz.write(output, snpdata3)
        snpdata4 = SnpNpz(output).read()
        assert snpdata3 == snpdata4

    def test_standardize_npz(self):
        snpreader = SnpNpz(self.currentFolder + "/examples/toydata10.snp.npz")
        self.standardize(snpreader)

    def test_c_reader_dat(self):
        snpreader = Dat(self.currentFolder + "/examples/toydata.dat")[:, ::100]
        _fortesting_JustCheckExists().input(snpreader)

        snpdata1 = snpreader.read()
        self.assertEqual(np.float64, snpdata1.val.dtype)
        self.assertTrue(
            np.allclose(self.snps[:, ::100], snpdata1.val, rtol=1e-05, atol=1e-05)
        )

        snpdata1.val[
            1, 2
        ] = np.NaN  # Inject a missing value to test writing and reading missing values
        output = "tempdir/snpreader/toydata.dat"
        create_directory_if_necessary(output)
        Dat.write(output, snpdata1)
        snpdata2 = Dat(output).read()
        np.testing.assert_array_almost_equal(snpdata1.val, snpdata2.val, decimal=10)

        snpdata3 = snpdata1[:, 0:0].read()  # create snpdata with no sids
        output = "tempdir/snpreader/toydata3.dat"
        Dat.write(output, snpdata3)
        snpdata4 = Dat(output).read()
        assert snpdata3 == snpdata4

    @staticmethod
    def assert_match_012_210(snpdata1, snpdata2):
        for sid_index in range(
            snpdata1.sid_count
        ):  # Check that every row matches (except OK if 0,1,2 can be 2,1,0)
            goods1 = (
                snpdata1.val[:, sid_index] == snpdata1.val[:, sid_index]
            )  # find non-missing
            goods2 = (
                snpdata2.val[:, sid_index] == snpdata2.val[:, sid_index]
            )  # find non-missing
            assert (goods1 == goods2).all()  # assert that they agree on non-missing
            is_ok = (
                snpdata1.val[goods1, sid_index] == snpdata2.val[goods2, sid_index]
            ).all() or (
                snpdata1.val[goods1, sid_index]
                == snpdata2.val[goods2, sid_index] * -1 + 2
            ).all()
            assert is_ok

    def test_c_reader_ped(self):
        if False:  # Too slow for routine testing
            snpdata1 = Ped(self.currentFolder + "/examples/toydata.ped")[
                ::25, ::1000
            ].read()
            self.assertEqual(np.float64, snpdata1.val.dtype)
            TestPySnpTools.assert_match_012_210(
                self.snpdata[::25, ::1000].read(), snpdata1
            )
        else:
            snpdata1 = self.snpdata[::25, ::1000].read()

        output = "tempdir/snpreader/toydata.ped"
        create_directory_if_necessary(output)

        snpdata1.val[
            1, 2
        ] = np.NaN  # Inject a missing value to test writing and reading missing values
        Ped.write(output, snpdata1)
        snpreader = Ped(output)
        _fortesting_JustCheckExists().input(snpreader)
        _ = str(snpreader)
        snpdata2 = snpreader.read()
        TestPySnpTools.assert_match_012_210(snpdata1, snpdata2)

    def test_c_reader_pheno(self):
        snpdata1 = Pheno(self.currentFolder + "/examples/toydata.phe").read()

        self.assertEqual(np.float64, snpdata1.val.dtype)

        snpdata1.val[
            1, 0
        ] = np.NaN  # Inject a missing value to test writing and reading missing values
        output = "tempdir/snpreader/toydata.phe"
        create_directory_if_necessary(output)
        Pheno.write(output, snpdata1)
        snpreader = Pheno(output)
        _fortesting_JustCheckExists().input(snpreader)
        _ = str(snpreader)
        snpdata2 = snpreader.read()
        np.testing.assert_array_almost_equal(snpdata1.val, snpdata2.val, decimal=10)

        snpdata1 = Pheno(self.currentFolder + "/examples/toydata.phe").read()
        import pysnptools.util.pheno as pstpheno

        dict = pstpheno.loadOnePhen(
            self.currentFolder + "/examples/toydata.phe", missing=""
        )
        snpdata3 = Pheno(dict).read()
        np.testing.assert_array_almost_equal(snpdata1.val, snpdata3.val, decimal=10)

        dict = pstpheno.loadOnePhen(
            self.currentFolder + "/examples/toydata.phe", missing="", vectorize=True
        )
        assert len(dict["vals"].shape) == 1, "test 1-d array of values"
        snpdata3 = Pheno(dict).read()
        np.testing.assert_array_almost_equal(snpdata1.val, snpdata3.val, decimal=10)

        snpdata4 = Pheno(None, iid_if_none=snpdata1.iid)
        assert (snpdata4.row == snpdata1.row).all() and snpdata4.col_count == 0

        snpdata5 = Pheno(self.currentFolder + "/examples/toydata.id.phe").read()
        np.testing.assert_array_almost_equal(snpdata1.val, snpdata5.val, decimal=10)
        snpdata6 = Pheno(self.currentFolder + "/examples/toydata.fid.phe").read()
        np.testing.assert_array_almost_equal(snpdata1.val, snpdata6.val, decimal=10)

    def test_c_reader_dense(self):
        snpdata1 = self.snpdata[:, ::100].read()
        snpdata1.val[
            1, 2
        ] = np.NaN  # Inject a missing value to test writing and reading missing values
        output = "tempdir/snpreader/toydata.dense.txt"
        create_directory_if_necessary(output)
        Dense.write(output, snpdata1)
        snpreader = Dense(output)
        _fortesting_JustCheckExists().input(snpreader)
        snpdata2 = snpreader.read()
        np.testing.assert_array_almost_equal(snpdata1.val, snpdata2.val, decimal=10)

    def test_c_reader_distributedbed(self):
        from pysnptools.util.filecache import LocalCache

        snpdata1 = self.snpdata[:, ::100].read()
        snpdata1.val[
            1, 2
        ] = np.NaN  # Inject a missing value to test writing and reading missing values
        output = "tempdir/snpreader/toydata.distributedbed"
        LocalCache(output).rmtree()
        DistributedBed.write(output, snpdata1, piece_per_chrom_count=5)
        snpreader = DistributedBed(output)
        _fortesting_JustCheckExists().input(snpreader)
        snpdata2 = snpreader.read()
        np.testing.assert_array_almost_equal(snpdata1.val, snpdata2.val, decimal=10)

    def test_some_std(self):
        k0 = self.snpdata.read_kernel(standardizer=Unit()).val

        k1 = self.snpdata.read_kernel(standardizer=Unit())
        np.testing.assert_array_almost_equal(k0, k1.val, decimal=10)

        from pysnptools.snpreader import SnpData

        snpdata2 = SnpData(
            iid=self.snpdata.iid,
            sid=self.snpdata.sid,
            pos=self.snpdata.pos,
            val=np.array(self.snpdata.val),
        )
        _ = str(snpdata2)
        snpdata2.standardize()
        _ = str(snpdata2)

        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        k2 = snpreader.read_kernel(standardizer=Unit(), block_size=500).val
        np.testing.assert_array_almost_equal(k0, k2, decimal=10)

        from pysnptools.standardizer.identity import Identity
        from pysnptools.standardizer.diag_K_to_N import DiagKtoN

        for dtype in [np.float64, np.float32]:
            for std in [Unit(), Beta(1, 25), Identity(), DiagKtoN()]:
                _ = str(std)
                np.random.seed(0)
                x = np.array(np.random.randint(3, size=[60, 100]), dtype=dtype)
                x2 = x[:, ::2]
                _ = np.array(x2)
                # LATER what's this about? It doesn't do non-contiguous?
                # assert not x2.flags['C_CONTIGUOUS'] and not x2.flags['F_CONTIGUOUS'] #set up to test non contiguous
                # assert x2b.flags['C_CONTIGUOUS'] or x2b.flags['F_CONTIGUOUS'] #set up to test non contiguous
                # a,b = std.standardize(x2b),std.standardize(x2)
                # np.testing.assert_array_almost_equal(a,b)

    def c_reader(self, snpreader):
        """
        make sure c-reader yields same result
        """
        snpdata = snpreader.read(order="F", force_python_only=False)
        snp_c = snpdata.val

        self.assertEqual(np.float64, snp_c.dtype)
        self.assertTrue(np.allclose(self.snps, snp_c, rtol=1e-05, atol=1e-05))
        return snpdata

    def test_standardize_bed(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        self.standardize(snpreader)

    def test_standardize_dat(self):
        snpreader = Dat(self.currentFolder + "/examples/toydata.dat")
        self.standardize(snpreader)

    def test_standardize_ped(self):
        snpreader = Ped(self.currentFolder + "/examples/toydata")
        self.standardize(snpreader)

    def standardize(self, snpreader):
        """
        make sure blocked standardize yields same result as regular standardize
        """

        for dtype in [np.float64, np.float32]:
            snps = snpreader.read(order="F", force_python_only=True, dtype=dtype).val
            self.assertEqual(dtype, snps.dtype)

            snp_s1 = Unit().standardize(snps.copy(), force_python_only=True)
            snp_s2 = Unit().standardize(
                snps.copy(), block_size=100, force_python_only=True
            )
            snps_F = np.array(snps, dtype=dtype, order="F")
            snp_s3 = Unit().standardize(snps_F)
            snps_C = np.array(snps, dtype=dtype, order="C")
            snp_s4 = Unit().standardize(snps_C)

            self.assertEqual(snp_s1.shape[0], snp_s2.shape[0])
            self.assertEqual(snp_s1.shape[1], snp_s2.shape[1])

            self.assertEqual(snp_s1.shape[0], snp_s3.shape[0])
            self.assertEqual(snp_s1.shape[1], snp_s3.shape[1])

            self.assertEqual(snp_s1.shape[0], snp_s4.shape[0])
            self.assertEqual(snp_s1.shape[1], snp_s4.shape[1])

            self.assertTrue(np.allclose(snp_s1, snp_s2, rtol=1e-05, atol=1e-05))
            self.assertTrue(np.allclose(snp_s1, snp_s3, rtol=1e-05, atol=1e-05))
            self.assertTrue(np.allclose(snp_s1, snp_s4, rtol=1e-05, atol=1e-05))

            snp_beta1 = Beta(1, 25).standardize(snps.copy(), force_python_only=True)
            snps_F = np.array(snps, dtype=dtype, order="F")
            snp_beta2 = Beta(1, 25).standardize(snps_F)
            snps_C = np.array(snps, dtype=dtype, order="C")
            snp_beta3 = Beta(1, 25).standardize(snps_C)

            self.assertEqual(snp_beta1.shape[0], snp_beta2.shape[0])
            self.assertEqual(snp_beta1.shape[1], snp_beta2.shape[1])
            self.assertEqual(snp_beta1.shape[0], snp_beta3.shape[0])
            self.assertEqual(snp_beta1.shape[1], snp_beta3.shape[1])

            self.assertTrue(np.allclose(snp_beta1, snp_beta2, rtol=1e-05, atol=1e-05))
            self.assertTrue(np.allclose(snp_beta1, snp_beta3, rtol=1e-05, atol=1e-05))

    def test_load_and_standardize_bed(self):
        snpreader2 = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        self.load_and_standardize(snpreader2, snpreader2)

        snpreader3 = Bed(
            self.currentFolder + "/examples/toydata.5chrom", count_A1=False
        )
        self.load_and_standardize(snpreader3, snpreader3)

    def too_slow_test_write_bedbig(self):
        iid_count = 100000
        sid_count = 50000
        from pysnptools.snpreader import SnpData

        iid = np.array([[str(i), str(i)] for i in range(iid_count)])
        sid = np.array(["sid_{0}".format(i) for i in range(sid_count)])
        pos = np.array([[i, i, i] for i in range(sid_count)])
        np.random.seed(0)
        snpdata = SnpData(
            iid, sid, np.zeros((iid_count, sid_count)), pos=pos
        )  # random.choice((0.0,1.0,2.0,float("nan")),size=(iid_count,sid_count)))
        output = "tempdir/bedbig.{0}.{1}".format(iid_count, sid_count)
        create_directory_if_necessary(output)
        Bed.write(output, snpdata, count_A1=False)
        snpdata2 = Bed(output, count_A1=False).read()
        np.testing.assert_array_almost_equal(snpdata.val, snpdata2.val, decimal=10)

    def test_write_bed_f64cpp_0(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        iid_index = 0
        logging.info("iid={0}".format(iid_index))
        # if snpreader.iid_count % 4 == 0: # divisible by 4 isn't a good test
        #    snpreader = snpreader[0:-1,:]
        # assert snpreader.iid_count % 4 != 0
        snpdata = snpreader[0:iid_index, :].read(order="F", dtype=np.float64)
        if snpdata.iid_count > 0:
            snpdata.val[-1, 0] = float("NAN")
        output = "tempdir/toydata.F64cpp.{0}.bed".format(iid_index)
        create_directory_if_necessary(output)
        Bed.write(output, snpdata, count_A1=False)
        snpdata2 = Bed(output, count_A1=False).read()
        np.testing.assert_array_almost_equal(snpdata.val, snpdata2.val, decimal=10)

    def test_write_bed_f64cpp_1(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        iid_index = 1
        logging.info("iid={0}".format(iid_index))
        # if snpreader.iid_count % 4 == 0: # divisible by 4 isn't a good test
        #    snpreader = snpreader[0:-1,:]
        # assert snpreader.iid_count % 4 != 0
        snpdata = snpreader[0:iid_index, :].read(order="F", dtype=np.float64)
        if snpdata.iid_count > 0:
            snpdata.val[-1, 0] = float("NAN")
        output = "tempdir/toydata.F64cpp.{0}".format(iid_index)
        create_directory_if_necessary(output)
        Bed.write(output, snpdata, count_A1=False)
        snpdata2 = Bed(output, count_A1=False).read()
        np.testing.assert_array_almost_equal(snpdata.val, snpdata2.val, decimal=10)

    def test_write_bed_f64cpp_5(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )

        from pysnptools.kernelreader.test import _fortesting_JustCheckExists

        _fortesting_JustCheckExists().input(snpreader)

        iid_index = 5
        logging.info("iid={0}".format(iid_index))
        # if snpreader.iid_count % 4 == 0: # divisible by 4 isn't a good test
        #    snpreader = snpreader[0:-1,:]
        # assert snpreader.iid_count % 4 != 0
        snpdata = snpreader[0:iid_index, :].read(order="F", dtype=np.float64)
        if snpdata.iid_count > 0:
            snpdata.val[-1, 0] = float("NAN")
        output = "tempdir/toydata.F64cpp.{0}".format(iid_index)
        create_directory_if_necessary(output)
        Bed.write(output, snpdata, count_A1=False)  # ,force_python_only=True)
        snpdata2 = Bed(output, count_A1=False).read()
        np.testing.assert_array_almost_equal(snpdata.val, snpdata2.val, decimal=10)

    def test_write_bed_f64cpp_5_python(self):
        snpreader = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        iid_index = 5
        logging.info("iid={0}".format(iid_index))
        # if snpreader.iid_count % 4 == 0: # divisible by 4 isn't a good test
        #    snpreader = snpreader[0:-1,:]
        # assert snpreader.iid_count % 4 != 0
        snpdata = snpreader[0:iid_index, :].read(order="F", dtype=np.float64)
        if snpdata.iid_count > 0:
            snpdata.val[-1, 0] = float("NAN")
        output = "tempdir/toydata.F64python.{0}".format(iid_index)
        create_directory_if_necessary(output)
        Bed.write(output, snpdata, force_python_only=True)
        snpdata2 = Bed(output, count_A1=False).read()
        np.testing.assert_array_almost_equal(snpdata.val, snpdata2.val, decimal=10)

    def test_write_x_x_cpp(self):
        for count_A1 in [False, True]:
            snpreader = Bed(
                self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=count_A1
            )
            for order in ["C", "F"]:
                for dtype in [np.float32, np.float64]:
                    snpdata = snpreader.read(order=order, dtype=dtype)
                    snpdata.val[-1, 0] = float("NAN")
                    output = "tempdir/toydata.{0}{1}.cpp".format(
                        order, "32" if dtype == np.float32 else "64"
                    )
                    create_directory_if_necessary(output)
                    Bed.write(output, snpdata, count_A1=count_A1)
                    snpdata2 = Bed(output, count_A1=count_A1).read()
                    np.testing.assert_array_almost_equal(
                        snpdata.val, snpdata2.val, decimal=10
                    )

    def test_subset_view(self):
        snpreader2 = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )[:, :]
        result = snpreader2.read(view_ok=True)
        self.assertFalse(snpreader2 is result)
        result2 = result[:, :].read()
        self.assertFalse(np.may_share_memory(result2.val, result.val))
        result3 = result[:, :].read(view_ok=True)
        self.assertTrue(np.may_share_memory(result3.val, result.val))
        result4 = result3.read()
        self.assertFalse(np.may_share_memory(result4.val, result3.val))
        result5 = result4.read(view_ok=True)
        self.assertTrue(np.may_share_memory(result4.val, result5.val))

    def test_load_and_standardize_hdf5(self):
        snpreader2 = SnpHdf5(self.currentFolder + "/examples/toydata.snpmajor.snp.hdf5")
        snpreader3 = SnpHdf5(self.currentFolder + "/examples/toydata.iidmajor.snp.hdf5")
        self.load_and_standardize(snpreader2, snpreader3)
        snpreaderref = Bed(
            self.currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        self.load_and_standardize(snpreader2, snpreaderref)

    def test_load_and_standardize_dat(self):
        snpreader2 = Dat(self.currentFolder + "/examples/toydata.dat")
        self.load_and_standardize(snpreader2, snpreader2)
        # snpreaderref = Bed(self.currentFolder + "/examples/toydata.5chrom.bed",count_A1=False)
        # self.load_and_standardize(snpreader2, snpreaderref)

    def test_load_and_standardize_ped(self):
        # !!Ped columns can be ambiguous
        # ##Creating Ped data ...
        # currentFolder = os.path.dirname(os.path.realpath(__file__))
        # snpData = Bed(currentFolder + "/examples/toydata.5chrom.bed",count_A1=False).read()
        # #Ped.write(snpData, currentFolder + "/examples/toydata.ped")
        # fromPed = Ped(currentFolder + "/examples/toydata").read()
        # self.assertTrue(np.allclose(snpData.val, fromPed.val, rtol=1e-05, atol=1e-05))

        snpreader2 = Ped(self.currentFolder + "/examples/toydata")
        self.load_and_standardize(snpreader2, snpreader2)
        # snpreaderref = Bed(self.currentFolder + "/examples/toydata.5chrom.bed",count_A1=False)
        # self.load_and_standardize(snpreader2, snpreaderref)

    def load_and_standardize(self, snpreader2, snpreader3):
        """
        test c-version of load and standardize
        """

        S = snpreader2.sid_count
        N_original = snpreader2.iid_count

        iid_index_list = range(N_original - 1, 0, -2)
        snpreader3 = snpreader3[iid_index_list, :]

        for dtype in [np.float64, np.float32]:
            G2 = snpreader2.read(order="F", force_python_only=True).val
            G2 = Unit().standardize(G2, block_size=10000, force_python_only=True)

            SNPs_floatF = snpreader2.read(
                order="F", dtype=dtype, force_python_only=False
            ).val
            GF = Unit().standardize(SNPs_floatF)

            SNPs_floatC = snpreader2.read(
                order="C", dtype=dtype, force_python_only=False
            ).val
            GC = Unit().standardize(SNPs_floatC)

            self.assertTrue(np.allclose(GF, G2, rtol=1e-05, atol=1e-05))
            self.assertTrue(np.allclose(GF, GC, rtol=1e-05, atol=1e-05))

            # testing selecting a subset of snps and iids
            snp_index_list = range(S - 1, 0, -2)

            G2x = snpreader2.read(order="F", force_python_only=True).val
            G2x = G2x[iid_index_list, :][:, snp_index_list]
            G2x = Unit().standardize(G2x, block_size=10000, force_python_only=True)

            SNPs_floatFx = (
                snpreader3[:, snp_index_list]
                .read(order="F", dtype=dtype, force_python_only=False)
                .val
            )
            GFx = Unit().standardize(SNPs_floatFx)
            self.assertTrue(np.allclose(GFx, G2x, rtol=1e-05, atol=1e-05))

            SNPs_floatCx = (
                snpreader3[:, snp_index_list]
                .read(order="C", dtype=dtype, force_python_only=False)
                .val
            )
            _ = Unit().standardize(SNPs_floatCx)
            self.assertTrue(np.allclose(GFx, G2x, rtol=1e-05, atol=1e-05))

    def test_val_is_float(self):
        from pysnptools.snpreader import SnpData
        from pysnptools.kernelreader import KernelData

        iid_count = 3
        sid_count = 2
        iid = [[str(i), str(i)] for i in range(iid_count)]
        sid = [str(s) for s in range(sid_count)]
        np.random.seed(0)
        vali = np.random.randint(10, size=[iid_count, sid_count])
        snpdata = SnpData(iid=iid, sid=sid, val=vali)
        assert snpdata.val.dtype == np.float64
        snpdata = SnpData(iid=iid, sid=sid, val=np.array(vali, dtype=np.float32))
        assert snpdata.val.dtype == np.float32
        snpdata = SnpData(iid=iid, sid=sid, val=np.array(vali, dtype=np.float64))
        assert snpdata.val.dtype == np.float64
        snpdata = SnpData(iid=iid, sid=sid, val=np.array(vali, dtype=np.float16))
        assert snpdata.val.dtype == np.float64
        assert (
            np.float64
            == SnpData(iid=iid, sid=sid, val=np.array(vali, dtype=np.str_)).val.dtype
        )

        valstr = np.array(vali, dtype=np.str_)
        valstr[0, 0] = "cannot convert"
        error_seen = False
        try:
            SnpData(iid=iid, sid=sid, val=valstr)  # expect error
        except Exception:
            error_seen = True
        assert error_seen
        error_seen = False
        try:
            KernelData(iid=iid, val=valstr)  # expect error
        except Exception:
            error_seen = True
        assert error_seen

    def test_read_dtype(self):
        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        bed = Bed("examples/toydata.5chrom.bed", count_A1=True)
        bed[:3, :2].read(dtype="float32").val

        os.chdir(previous_wd)

    def test_respect_read_inputs(self):
        from pysnptools.snpreader import _MergeIIDs, _MergeSIDs, SnpGen, SnpMemMap

        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        snpreader_list = [
            Bed("examples/toydata.5chrom.bed", count_A1=True),
            SnpGen(seed=0, iid_count=500, sid_count=50),
            SnpHdf5("examples/toydata.snpmajor.snp.hdf5"),
            SnpMemMap("examples/tiny.snp.memmap"),
            SnpNpz("examples/toydata10.snp.npz"),
            Bed("examples/toydata.5chrom.bed", count_A1=True)[::2, ::2],
            _MergeSIDs(
                [
                    Bed("examples/toydata.5chrom.bed", count_A1=True)[:, :5].read(),
                    Bed("examples/toydata.5chrom.bed", count_A1=True)[:, 5:].read(),
                ]
            ),
            Dat("examples/toydata.dat"),
            Dense("examples/toydata100.dense.txt"),
            DistributedBed("examples/toydataSkip10.distributedbed"),
            Ped("examples/toydata.ped"),
            Pheno("examples/toydata.phe"),
            Bed("examples/toydata.5chrom.bed", count_A1=True).read(),
            _MergeIIDs(
                [
                    Bed("examples/toydata.5chrom.bed", count_A1=True)[:5, :].read(),
                    Bed("examples/toydata.5chrom.bed", count_A1=True)[5:, :].read(),
                ]
            ),
        ]

        output_p = "tempdir/snpreader/respect.p"
        create_directory_if_necessary(output_p)

        for snpreader in snpreader_list:
            logging.info(str(snpreader))
            for order in ["F", "C", "A"]:
                for dtype in [np.float32, np.float64]:
                    for force_python_only in [True, False]:
                        for view_ok in [True, False]:
                            val = snpreader.read(
                                order=order,
                                dtype=dtype,
                                force_python_only=force_python_only,
                                view_ok=view_ok,
                            ).val
                            has_right_order = (
                                order == "A"
                                or (order == "C" and val.flags["C_CONTIGUOUS"])
                                or (order == "F" and val.flags["F_CONTIGUOUS"])
                            )
                            if hasattr(snpreader, "val") and not view_ok:
                                assert snpreader.val is not val
                            if (
                                hasattr(snpreader, "val")
                                and view_ok
                                and snpreader.val is not val
                                and (
                                    order == "A"
                                    or (
                                        order == "F"
                                        and snpreader.val.flags["F_CONTIGUOUS"]
                                    )
                                    or (
                                        order == "C"
                                        and snpreader.val.flags["C_CONTIGUOUS"]
                                    )
                                )
                                and (dtype is None or snpreader.val.dtype == dtype)
                            ):
                                logging.info(
                                    "{0} could have read a view, but didn't".format(
                                        snpreader
                                    )
                                )
                            if not force_python_only:
                                # Don't check this when force_python_only -- Bed is know to get the order wrong,
                                # but it doesn't matter
                                assert val.dtype == dtype and has_right_order
            with open(output_p, "wb") as f:
                pickle.dump(snpreader, f)
            with open(output_p, "rb") as f:
                snpreader_p = pickle.load(f)
            val_p = snpreader_p.read(
                order=order,
                dtype=dtype,
                force_python_only=force_python_only,
                view_ok=view_ok,
            ).val
            np.allclose(val, val_p, equal_nan=True)

        os.chdir(previous_wd)

    def test_writes(self):
        from pysnptools.snpreader import SnpData, SnpMemMap

        the_class_and_suffix_list = [(SnpMemMap, "memmap", None, None)]

        # ===================================
        #    Starting main function
        # ===================================
        logging.info("starting 'test_writes'")
        np.random.seed(0)
        output_template = "tempdir/snpreader/writes.{0}.{1}"
        create_directory_if_necessary(output_template.format(0, "npz"))
        for row_count in [1]:
            for col_count in [4]:
                val = np.random.randint(0, 4, size=(row_count, col_count)) * 1.0
                val[val == 3] = np.NaN
                row = [("0", "0"), ("1", "1"), ("2", "2"), ("3", "3"), ("4", "4")][
                    :row_count
                ]
                col = ["s0", "s1", "s2", "s3", "s4"][:col_count]
                for is_none in [True]:
                    col_prop = None
                    snpdata = SnpData(iid=row, sid=col, val=val, pos=col_prop)
                    for (
                        the_class,
                        suffix,
                        constructor,
                        writer,
                    ) in the_class_and_suffix_list:

                        def constructor(filename):
                            return the_class(filename)

                        def writer(filename, _data):
                            return the_class.write(filename, _data)

                        filename = output_template.format("debug", suffix)
                        logging.info(filename)
                        ret = writer(filename, snpdata)
                        assert ret is not None
                        for subsetter in [np.s_[::2, ::3]]:
                            reader = constructor(filename)
                            _fortesting_JustCheckExists().input(reader)
                            subreader = (
                                reader
                                if subsetter is None
                                else reader[subsetter[0], subsetter[1]]
                            )
                            _ = subreader.read(order="C")
                        try:
                            os.remove(filename)
                        except Exception:
                            pass
        logging.info("done with 'test_writes'")

    def test_writes2(self):
        from pysnptools.snpreader import SnpData, SnpHdf5, SnpNpz, SnpMemMap

        the_class_and_suffix_list = [
            (DistributedBed, "distributed_bed", None, None),
            (Dense, "dense", None, None),
            (Bed, "bed", lambda filename: Bed(filename, count_A1=False), None),
            (Dat, "dat", None, None),
            (Ped, "ped", None, None),
            (Pheno, "pheno", None, None),
            (SnpHdf5, "hdf5", None, None),
            (SnpNpz, "npz", None, None),
            (SnpMemMap, "memmap", None, None),
        ]
        cant_do_col_prop_none_set = {"dense", "distributed_bed"}
        cant_do_col_len_0_set = {"distributed_bed"}
        cant_do_row_count_zero_set = {"dense", "ped", "pheno"}
        can_swap_0_2_set = {"ped"}
        can_change_col_names_set = {"pheno"}
        ignore_fam_id_set = {"dense"}
        ignore_pos_set = {"dense", "pheno"}
        erase_any_write_dir = {"distributed_bed"}

        # ===================================
        #    Starting main function
        # ===================================
        logging.info("starting 'test_writes'")
        np.random.seed(0)
        output_template = "tempdir/snpreader/writes.{0}.{1}"
        create_directory_if_necessary(output_template.format(0, "npz"))
        i = 0
        for row_count in [0, 5, 2, 1]:
            for col_count in [4, 2, 1, 0]:
                val = np.random.randint(0, 4, size=(row_count, col_count)) * 1.0
                val[val == 3] = np.NaN
                row = [("0", "0"), ("1", "1"), ("2", "2"), ("3", "3"), ("4", "4")][
                    :row_count
                ]
                col = ["s0", "s1", "s2", "s3", "s4"][:col_count]
                for is_none in [True, False]:
                    if is_none:
                        col_prop = None
                    else:
                        col_prop = [(x, x, x) for x in range(1, 6)][:col_count]
                    snpdata = SnpData(
                        iid=row, sid=col, val=val, pos=col_prop, name=str(i)
                    )
                    for (
                        the_class,
                        suffix,
                        constructor,
                        writer,
                    ) in the_class_and_suffix_list:
                        constructor = constructor or (
                            lambda filename: the_class(filename)
                        )
                        writer = writer or (
                            lambda filename, _data: the_class.write(filename, _data)
                        )

                        if col_count == 0 and suffix in cant_do_col_len_0_set:
                            continue
                        if col_prop is None and suffix in cant_do_col_prop_none_set:
                            continue
                        if row_count == 0 and suffix in cant_do_row_count_zero_set:
                            continue
                        filename = output_template.format(i, suffix)
                        logging.info(filename)
                        i += 1
                        if suffix in erase_any_write_dir and os.path.exists(filename):
                            shutil.rmtree(filename)
                        ret = writer(filename, snpdata)
                        assert ret is not None
                        for subsetter in [None, np.s_[::2, ::3]]:
                            reader = constructor(filename)
                            _fortesting_JustCheckExists().input(reader)
                            subreader = (
                                reader
                                if subsetter is None
                                else reader[subsetter[0], subsetter[1]]
                            )
                            readdata = subreader.read(order="C")
                            expected = (
                                snpdata
                                if subsetter is None
                                else snpdata[subsetter[0], subsetter[1]].read()
                            )
                            if suffix not in can_swap_0_2_set:
                                assert np.allclose(
                                    readdata.val, expected.val, equal_nan=True
                                )
                            else:
                                for col_index in range(readdata.col_count):
                                    assert np.allclose(
                                        readdata.val[:, col_index],
                                        expected.val[:, col_index],
                                        equal_nan=True,
                                    ) or np.allclose(
                                        readdata.val[:, col_index] * -1 + 2,
                                        expected.val[:, col_index],
                                        equal_nan=True,
                                    )
                            if suffix not in ignore_fam_id_set:
                                assert np.array_equal(readdata.row, expected.row)
                            else:
                                assert np.array_equal(
                                    readdata.row[:, 1], expected.row[:, 1]
                                )
                            if suffix not in can_change_col_names_set:
                                assert np.array_equal(readdata.col, expected.col)
                            else:
                                assert readdata.col_count == expected.col_count
                            assert np.array_equal(
                                readdata.row_property, expected.row_property
                            ) or (
                                readdata.row_property.shape[1] == 0
                                and expected.row_property.shape[1] == 0
                            )

                            if suffix not in ignore_pos_set:
                                assert np.allclose(
                                    readdata.col_property,
                                    expected.col_property,
                                    equal_nan=True,
                                ) or (
                                    readdata.col_property.shape[1] == 0
                                    and expected.col_property.shape[1] == 0
                                )
                            else:
                                assert len(readdata.col_property) == len(
                                    expected.col_property
                                )
                        try:
                            os.remove(filename)
                        except Exception:
                            pass
        logging.info("done with 'test_writes'")


class NaNCNCTestCases(unittest.TestCase):
    def __init__(
        self,
        iid_index_list,
        snp_index_list,
        standardizer,
        snpreader,
        dtype,
        order,
        force_python_only,
        reference_snps,
        reference_dtype,
    ):
        self.iid_index_list = iid_index_list
        self.snp_index_list = snp_index_list
        self.standardizer = standardizer
        self.snpreader = snpreader
        self.dtype = dtype
        self.order = order
        self.force_python_only = force_python_only
        self.reference_snps = reference_snps
        self.reference_dtype = reference_dtype

    _testMethodName = "runTest"
    _testMethodDoc = None

    @staticmethod
    def factory_iterator():
        def snp_reader_factory_bed():
            return Bed("examples/toydata.5chrom.bed", count_A1=False)

        def snp_reader_factory_snpmajor_hdf5():
            return SnpHdf5("examples/toydata.snpmajor.snp.hdf5")

        def snp_reader_factory_iidmajor_hdf5():
            return SnpHdf5("examples/toydata.iidmajor.snp.hdf5")

        def snp_reader_factory_dat():
            return Dat("examples/toydata.dat")

        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        snpreader0 = snp_reader_factory_bed()
        S_original = snpreader0.sid_count
        N_original = snpreader0.iid_count

        snps_to_read_count = min(S_original, 100)

        for iid_index_list in [
            range(N_original),
            range(N_original // 2),
            range(N_original - 1, 0, -2),
        ]:
            for snp_index_list in [
                range(snps_to_read_count),
                range(snps_to_read_count // 2),
                range(snps_to_read_count - 1, 0, -2),
            ]:
                for standardizer in [Unit(), Beta(1, 25)]:
                    reference_snps, reference_dtype = NaNCNCTestCases(
                        iid_index_list,
                        snp_index_list,
                        standardizer,
                        snp_reader_factory_bed(),
                        np.float64,
                        "C",
                        "False",
                        None,
                        None,
                    ).read_and_standardize()
                    for snpreader_factory in [
                        snp_reader_factory_bed,
                        snp_reader_factory_snpmajor_hdf5,
                        snp_reader_factory_iidmajor_hdf5,
                        snp_reader_factory_dat,
                    ]:
                        for dtype in [np.float64, np.float32]:
                            for order in ["C", "F"]:
                                for force_python_only in [False, True]:
                                    snpreader = snpreader_factory()
                                    test_case = NaNCNCTestCases(
                                        iid_index_list,
                                        snp_index_list,
                                        standardizer,
                                        snpreader,
                                        dtype,
                                        order,
                                        force_python_only,
                                        reference_snps,
                                        reference_dtype,
                                    )
                                    yield test_case
        os.chdir(previous_wd)

    def __str__(self):
        iid_index_list = self.iid_index_list
        snp_index_list = self.snp_index_list
        standardizer = self.standardizer
        snpreader = self.snpreader
        dtype = self.dtype
        order = self.order
        force_python_only = self.force_python_only
        return (
            "{0}(iid_index_list=[{1}], snp_index_list=[{2}], standardizer={3}, snpreader={4}, "
            + "dtype={5}, order='{6}', force_python_only=={7})"
        ).format(
            self.__class__.__name__,
            ",".join([str(i) for i in iid_index_list])
            if len(iid_index_list) < 10
            else ",".join([str(i) for i in iid_index_list[0:10]]) + ",...",
            ",".join([str(i) for i in snp_index_list])
            if len(snp_index_list) < 10
            else ",".join([str(i) for i in snp_index_list[0:10]]) + ",...",
            standardizer,
            snpreader,
            dtype,
            order,
            force_python_only,
        )

    def read_and_standardize(self):
        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        iid_index_list = self.iid_index_list
        snp_index_list = self.snp_index_list
        standardizer = self.standardizer
        snpreader = self.snpreader
        dtype = self.dtype
        order = self.order
        force_python_only = self.force_python_only

        snps = (
            snpreader[iid_index_list, snp_index_list]
            .read(order=order, dtype=dtype, force_python_only=force_python_only)
            .val
        )
        snps[0, 0] = np.nan  # add a NaN
        snps[:, 1] = 2  # make a SNC
        snps = standardizer.standardize(snps, force_python_only=force_python_only)
        os.chdir(previous_wd)
        return snps, dtype

    def doCleanups(self):
        pass
        # return super(NaNCNCTestCases, self).doCleanups()

    def runTest(self, result=None):
        snps, dtype = self.read_and_standardize()
        self.assertTrue(snps[0, 0] == 0)
        self.assertTrue(np.all(snps[:, 1] == 0))
        if self.reference_snps is not None:
            self.assertTrue(
                np.allclose(
                    self.reference_snps,
                    snps,
                    rtol=1e-04
                    if dtype == np.float32 or self.reference_dtype == np.float32
                    else 1e-12,
                )
            )


# We do it this way instead of using doctest.DocTestSuite because doctest.DocTestSuite requires modules to be pickled,
# which python doesn't allow. We need tests to be pickleable so that they can be run on a cluster.
class TestSnpDocStrings(unittest.TestCase):
    def test_bed(self):
        import pysnptools.snpreader.bed

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        result = doctest.testmod(pysnptools.snpreader.bed, optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_dat(self):
        import pysnptools.snpreader.snpdata

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        result = doctest.testmod(pysnptools.snpreader.dat, optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_dense(self):
        import pysnptools.snpreader.snpdata

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        result = doctest.testmod(
            pysnptools.snpreader.dense, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_pairs(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        import pysnptools.snpreader.pairs

        result = doctest.testmod(
            pysnptools.snpreader.pairs, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_ped(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        import pysnptools.snpreader.ped

        result = doctest.testmod(pysnptools.snpreader.ped, optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_pheno(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        import pysnptools.snpreader.pheno

        result = doctest.testmod(
            pysnptools.snpreader.pheno, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_snpdata(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        import pysnptools.snpreader.snpdata

        result = doctest.testmod(
            pysnptools.snpreader.snpdata, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_snphdf5(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        import pysnptools.snpreader.snphdf5

        result = doctest.testmod(
            pysnptools.snpreader.snphdf5, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_snpmemmap(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        import pysnptools.snpreader.snpmemmap

        result = doctest.testmod(
            pysnptools.snpreader.snpmemmap, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_snpgen(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/util")
        import pysnptools.snpreader.snpgen

        result = doctest.testmod(
            pysnptools.snpreader.snpgen, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_snpnpz(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        import pysnptools.snpreader.snpnpz

        result = doctest.testmod(
            pysnptools.snpreader.snpnpz, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_distributedbed(self):
        import pysnptools.snpreader.distributedbed

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        result = doctest.testmod(
            pysnptools.snpreader.distributedbed, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_snpreader(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/snpreader")
        import pysnptools.snpreader.snpreader

        result = doctest.testmod(
            pysnptools.snpreader.snpreader,
            optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE,
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_util(self):
        import pysnptools.util

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/util")

        result = doctest.testmod(pysnptools.util, optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_standardize_testmod(self):
        import pysnptools.standardizer

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/standardizer")
        for mod in [
            pysnptools.standardizer.beta,
            pysnptools.standardizer.betatrained,
            pysnptools.standardizer.bysidcount,
            pysnptools.standardizer.bysqrtsidcount,
            pysnptools.standardizer.diag_K_to_N,
            pysnptools.standardizer.identity,
            pysnptools.standardizer.standardizer,
            pysnptools.standardizer.unit,
            pysnptools.standardizer.unittrained,
        ]:
            result = doctest.testmod(
                mod, optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE
            )
            assert result.failed == 0, "failed doc test: " + __file__
        os.chdir(old_dir)

    def test_kernelstandardize_testmod(self):
        import pysnptools.kernelstandardizer

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/kernelstandardizer")
        for mod in [
            pysnptools.kernelstandardizer,
        ]:
            result = doctest.testmod(mod)
            assert result.failed == 0, "failed doc test: " + __file__
        os.chdir(old_dir)


def getTestSuite():
    """
    set up composite test suite
    """
    test_suite = unittest.TestSuite([])

    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistReaders))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPySnpTools))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistributedBed))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestFileCache))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestUtilTools))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestIntRangeSet))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSnpDocStrings))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPstDocStrings))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestKrDocStrings))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSnpGen))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestGenerate))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestExampleFile))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPstMemMap))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSnpMemMap))
    test_suite.addTests(NaNCNCTestCases.factory_iterator())
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPstReader))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestKernelReader))

    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    ret = r.run(suites)
    assert ret.wasSuccessful()



# !!!move these to their own files

import numpy as np
import logging


class KernelStandardizer(object):
    """
    A KernelStandardizer is a class such as :class:`.DiagKtoN` and :class:`.Identity` to be used by the :meth:`.KernelData.standardize` to standardize Kernel data.
    It always works in-place *and* returns the :class:`.KernelData` on which it works.

    :Example:

    Read and standardize Kernel data.

    >>> from pysnptools.kernelstandardizer import DiagKtoN
    >>> from pysnptools.kernelreader import KernelNpz
    >>> from pysnptools.util import example_file # Download and return local file name
    >>>
    >>> kernel_file = example_file('pysnptools/examples/toydata.kernel.npz')
    >>> kerneldata1 = KernelNpz(kernel_file).read()
    >>> print(np.diag(kerneldata1.val).sum())
    5000000.0
    >>> kerneldata1 = kerneldata1.standardize(DiagKtoN())
    >>> print(np.diag(kerneldata1.val).sum())
    500.0

    Can also return a constant kernel standardizer that be applied to other :class:`.KernelData`.

    >>> kernel_whole = KernelNpz(kernel_file)
    >>> train_idx, test_idx = range(10,kernel_whole.iid_count), range(0,10)  #test on the first 10, train on the rest
    >>> kernel_train, trained_standardizer = DiagKtoN().standardize(kernel_whole[train_idx].read(),return_trained=True)
    >>> print('{0:.6f}'.format(np.diag(kernel_train.val).sum()))
    490.000000
    >>> print('{0:.6f}'.format(trained_standardizer.factor))
    0.000100
    >>> kernel_whole_test = kernel_whole[:,test_idx].read().standardize(trained_standardizer)
    >>> print('{0:.6f}'.format(kernel_whole_test.val[0,0]))
    0.992217


    Details of Methods & Properties:
    """

    def standardize(
        self,
        kerneldata,
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):
        """
        Applies standardization, in place, to :class:`.KernelData`. For convenience also returns the :class:`KernelData`.

        :param snps: kernel values to standardize
        :type snps: :class:`.KernelData`

        :param return_trained: If true, returns a second value containing a constant :class:`.KernelStandardizer` trained on this data.
        :type return_trained: bool

        :param force_python_only: optional -- If False (default), may use outside library code. If True, requests that the read
            be done without outside library code.
        :type force_python_only: bool

        :param num_threads: optional -- The number of threads with which to standardize data. Defaults to all available
            processors. Can also be set with these environment variables (listed in priority order):
            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
        :type num_threads: None or int


        :rtype: :class:`.KernelData`, (optional) constant :class:`.KernelStandardizer`

        """
        raise NotImplementedError(
            "subclass {0} needs to implement method '.standardize'".format(
                self.__class__.__name__
            )
        )


class Identity(KernelStandardizer):
    """
    A :class:`.KernelStandardizer` that does nothing to kernel data.

    See :class:`.KernelStandardizer` for more information about standardization.

    >>> from pysnptools.kernelstandardizer import Identity as KS_Identity
    >>> from pysnptools.kernelreader import KernelNpz
    >>> from pysnptools.util import example_file # Download and return local file name
    >>>
    >>> kernel_file = example_file('pysnptools/examples/toydata.kernel.npz')
    >>> kerneldata1 = KernelNpz(kernel_file).read()
    >>> print(np.diag(kerneldata1.val).sum())
    5000000.0
    >>> kerneldata1 = kerneldata1.standardize(KS_Identity())
    >>> print(np.diag(kerneldata1.val).sum())
    5000000.0
    """

    def __init__(self):
        super(Identity, self).__init__()

    def standardize(
        self,
        kerneldata,
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):
        if return_trained:
            return kerneldata, self
        else:
            return kerneldata

    def __repr__(self):
        return "{0}()".format(self.__class__.__name__)


from pysnptools.standardizer import DiagKtoN  # as SN_DiagKtoN
from pysnptools.standardizer import DiagKtoNTrained  # as SN_DiagKtoNTrained


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from pysnptools.util.mapreduce1.runner import *
import os
import subprocess, sys, os.path
import multiprocessing
import pysnptools.util as pstutil
import pdb
import logging
import cloudpickle as pickle


class HPC(Runner):
    """
    Old code to run on a Microsoft Widows HPC Cluster. Not currently supported.
    """

    # !!LATER make it (and Hadoop) work from root directories -- or give a clear error message
    def __init__(
        self,
        taskcount,
        clustername,
        fileshare,
        priority="Normal",
        unit="core",
        mkl_num_threads=None,
        runtime="infinite",
        remote_python_parent=None,
        update_remote_python_parent=False,
        min=None,
        max=None,
        excluded_nodes=[],
        template=None,
        nodegroups=None,
        skipinputcopy=False,
        node_local=True,
        clean_up=True,
        preemptable=True,
        FailOnTaskFailure=False,
        logging_handler=logging.StreamHandler(sys.stdout),
    ):
        logger = logging.getLogger()
        if not logger.handlers:
            logger.setLevel(logging.INFO)
        for h in list(logger.handlers):
            logger.removeHandler(h)
        logger.addHandler(logging_handler)
        if logger.level == logging.NOTSET:
            logger.setLevel(logging.INFO)

        self.taskcount = taskcount
        self.clustername = clustername
        self.fileshare = fileshare

        self.priority = priority
        self.runtime = runtime
        self.unit = unit
        self.excluded_nodes = excluded_nodes
        self.min = min
        self.max = max
        self.remote_python_parent = remote_python_parent
        self.update_remote_python_parent = update_remote_python_parent
        self.CheckUnitAndMKLNumThreads(mkl_num_threads, unit)
        self.skipinputcopy = skipinputcopy
        self.template = template
        self.nodegroups = nodegroups
        self.node_local = node_local
        self.clean_up = clean_up
        self.preemptable = preemptable
        self.FailOnTaskFailure = FailOnTaskFailure

    def run(self, distributable):
        # Check that the local machine has python path set
        localpythonpath = os.environ.get(
            "PYTHONPATH"
        )  # !!should it be able to work without pythonpath being set (e.g. if there was just one file)? Also, is None really the return or is it an exception.
        if localpythonpath is None:
            raise Exception("Expect local machine to have 'pythonpath' set")

        remotepythoninstall = self.check_remote_pythoninstall()

        remotewd, run_dir_abs, run_dir_rel, nodelocalwd = self.create_run_dir()
        pstutil.create_directory_if_necessary(
            os.path.join(remotewd, distributable.tempdirectory), isfile=False
        )  # create temp directory now so that cluster tasks won't try to create it many times at once
        result_remote = os.path.join(run_dir_abs, "result.p")

        self.copy_python_settings(run_dir_abs)

        inputOutputCopier = HPCCopier(
            remotewd, skipinput=self.skipinputcopy
        )  # Create the object that copies input and output files to where they are needed

        inputOutputCopier.input(
            distributable
        )  # copy of the input files to where they are needed (i.e. the cluster)

        remotepythonpath = self.FindOrCreateRemotePythonPath(
            localpythonpath, run_dir_abs
        )

        batfilename_rel = self.create_bat_file(
            distributable,
            remotepythoninstall,
            remotepythonpath,
            remotewd,
            run_dir_abs,
            run_dir_rel,
            result_remote,
            nodelocalwd,
            distributable,
        )

        self.submit_to_cluster(
            batfilename_rel,
            distributable,
            remotewd,
            run_dir_abs,
            run_dir_rel,
            nodelocalwd,
        )

        inputOutputCopier.output(
            distributable
        )  # copy the output file from where they were created (i.e. the cluster) to the local computer

        assert os.path.exists(
            result_remote
        ), "The HPC job produced no result (and, thus, likely failed)"
        with open(result_remote, mode="rb") as f:
            result = pickle.load(f)

        # logging.info('Done: HPC runner is running a distributable. Returns {0}'.format(result))
        return result

    def CheckUnitAndMKLNumThreads(self, mkl_num_threads, unit):
        if unit.lower() == "core":
            if mkl_num_threads is not None and mkl_num_threads != 1:
                raise Exception(
                    "When 'unit' is 'core', mkl_num_threads must be unspecified or 1"
                )
            self.mkl_num_threads = 1
        elif unit.lower() == "socket":
            if mkl_num_threads is None:
                raise Exception(
                    "When 'unit' is 'socket', mkl_num_threads must be specified"
                )
            self.mkl_num_threads = mkl_num_threads
        elif unit.lower() == "node":
            self.mkl_num_threads = mkl_num_threads
        else:
            raise Exception("Expect 'unit' to be 'core', 'socket', or 'node'")

    def copy_python_settings(self, run_dir_abs):
        # localuserprofile = os.environ.get("USERPROFILE")
        user_python_settings = ".continuum"
        python_settings = os.path.join(self.fileshare, user_python_settings)
        if os.path.exists(python_settings):
            import shutil

            remote_user_python_settings = os.path.join(
                run_dir_abs, user_python_settings
            )
            shutil.copytree(python_settings, remote_user_python_settings)

    def FindOrCreateRemotePythonPath(self, localpythonpath, run_dir_abs):
        if self.remote_python_parent is None:
            remotepythonpath = self.CopySource(localpythonpath, run_dir_abs)
        else:
            pstutil.create_directory_if_necessary(
                self.remote_python_parent, isfile=False
            )
            list = []
            for rel in os.listdir(self.remote_python_parent):
                list.append(os.path.join(self.remote_python_parent, rel))
            remotepythonpath = ";".join(list)
            if self.update_remote_python_parent:
                remotepythonpath = self.CopySource(localpythonpath, run_dir_abs)

        return remotepythonpath

    def numString(self):
        if self.min is None and self.max is None:
            return " -Num{0} *-*".format(self.unit.capitalize())
        if self.min is None:
            return " -Num{0} {1}".format(self.unit.capitalize(), self.max)
        if self.max is None:
            return " -Num{0} {1}-*".format(self.unit.capitalize(), self.min)
        return " -Num{0} {1}-{2}".format(self.unit.capitalize(), self.min, self.max)

    def submit_to_cluster(
        self,
        batfilename_rel,
        distributable,
        remotewd,
        run_dir_abs,
        run_dir_rel,
        nodelocalwd,
    ):
        stdout_dir_rel = os.path.join(run_dir_rel, "stdout")
        stdout_dir_abs = os.path.join(run_dir_abs, "stdout")
        pstutil.create_directory_if_necessary(stdout_dir_abs, isfile=False)
        stderr_dir_rel = os.path.join(run_dir_rel, "stderr")
        stderr_dir_abs = os.path.join(run_dir_abs, "stderr")
        pstutil.create_directory_if_necessary(stderr_dir_abs, isfile=False)

        if len(self.excluded_nodes) > 0:
            excluded_nodes = "Set-HpcJob -Id $r.Id -addExcludedNodes {0}".format(
                ", ".join(self.excluded_nodes)
            )
        else:
            excluded_nodes = ""

        # create the Powershell file
        psfilename_rel = os.path.join(run_dir_rel, "dist.ps1")
        psfilename_abs = os.path.join(run_dir_abs, "dist.ps1")
        pstutil.create_directory_if_necessary(psfilename_abs, isfile=True)
        with open(psfilename_abs, "w") as psfile:
            psfile.write(
                r"""Add-PsSnapin Microsoft.HPC
        Set-Content Env:CCP_SCHEDULER {0}
        $r = New-HpcJob -Name "{7}" -Priority {8}{12}{14}{16} -RunTime {15} -FailOnTaskFailure {23} #-Preemptable {22}
        $r.Id
        if ({20})
        {10}
            $from = "{4}"
            $to = "{17}"
            Add-HpcTask -Name NodePrep    -JobId $r.Id -Type NodePrep                -CommandLine "${{from}}\{18}"        -StdOut "${{from}}\{2}\nodeprep.txt"    -StdErr "${{from}}\{3}\nodeprep.txt"    -WorkDir .
            Add-HpcTask -Name Parametric  -JobId $r.Id -Parametric -Start 0 -End {1} -CommandLine "${{from}}\{6} * {5}"   -StdOut "${{from}}\{2}\*.txt"    -StdErr "${{from}}\{3}\*.txt"                  -WorkDir $to
            Add-HpcTask -Name Reduce      -JobId $r.Id -Depend Parametric            -CommandLine "${{from}}\{6} {5} {5}" -StdOut "${{from}}\{2}\reduce.txt"      -StdErr "${{from}}\{3}\reduce.txt"      -WorkDir $to
            {21}Add-HpcTask -Name NodeRelease -JobId $r.Id -Type NodeRelease         -CommandLine "${{from}}\{19}"        -StdOut "${{from}}\{2}\noderelease.txt" -StdErr "${{from}}\{3}\noderelease.txt" -WorkDir .
        {11}
        else
        {10}
            Add-HpcTask -Name Parametric -JobId $r.Id -Parametric -Start 0 -End {1} -CommandLine "{6} * {5}" -StdOut "{2}\*.txt" -StdErr "{3}\*.txt" -WorkDir {4}
            Add-HpcTask -Name Reduce -JobId $r.Id -Depend Parametric -CommandLine "{6} {5} {5}" -StdOut "{2}\reduce.txt" -StdErr "{3}\reduce.txt" -WorkDir {4}
        {11}

        {13}
        Submit-HpcJob -Id $r.Id
        $j = Get-HpcJob -Id $r.Id
        $i = $r.id
        $s = 10

        while(($j.State -ne "Finished") -and ($j.State -ne "Failed") -and ($j.State -ne "Canceled"))
        {10}
            $x = $j.State
            Write-Host "${10}x{11}. Job# ${10}i{11} sleeping for ${10}s{11}"
            Start-Sleep -s $s
            if ($s -ge 60)
            {10}
            $s = 60
            {11}
            else
            {10}
                $s = $s * 1.1
            {11}
           $j.Refresh()
        {11}

        """.format(
                    self.clustername,  # 0
                    self.taskcount - 1,  # 1
                    stdout_dir_rel,  # 2
                    stderr_dir_rel,  # 3
                    remotewd,  # 4 fileshare wd
                    self.taskcount,  # 5
                    batfilename_rel,  # 6
                    self.maxlen(str(distributable), 50),  # 7
                    self.priority,  # 8
                    self.unit,  # 9 -- not used anymore,. Instead #12 sets unit
                    "{",  # 10
                    "}",  # 11
                    self.numString(),  # 12
                    excluded_nodes,  # 13
                    ' -templateName "{0}"'.format(self.template)
                    if self.template is not None
                    else "",  # 14
                    self.runtime,  # 15 RuntimeSeconds
                    ' -NodeGroups "{0}"'.format(self.nodegroups)
                    if self.nodegroups is not None
                    else "",  # 16
                    nodelocalwd,  # 17 the node-local wd
                    batfilename_rel[0:-8] + "nodeprep.bat",  # 18
                    batfilename_rel[0:-8] + "noderelease.bat",  # 19
                    1 if self.node_local else 0,  # 20
                    "",  # 21 always run release task
                    self.preemptable,  # 22
                    "$true" if self.FailOnTaskFailure else "$false",  # 23
                )
            )
        assert batfilename_rel[-8:] == "dist.bat", "real assert"
        import subprocess

        proc = subprocess.Popen(
            ["powershell.exe", "-ExecutionPolicy", "Unrestricted", psfilename_abs],
            cwd=os.getcwd(),
        )
        if not 0 == proc.wait():
            raise Exception(
                "Running powershell cluster submit script results in non-zero return code"
            )

    # move to utils?
    @staticmethod
    def maxlen(s, max):
        """
        Truncate cluster job name if longer than max.
        """
        if len(s) <= max:
            return s
        else:
            # return s[0:max-1]
            return s[-max:]  # JL: I prefer the end of the name rather than the start

    def create_distributablep(self, distributable, run_dir_abs, run_dir_rel):
        distributablep_filename_rel = os.path.join(run_dir_rel, "distributable.p")
        distributablep_filename_abs = os.path.join(run_dir_abs, "distributable.p")
        with open(distributablep_filename_abs, mode="wb") as f:
            pickle.dump(distributable, f)
        return distributablep_filename_rel, distributablep_filename_abs

    @staticmethod
    def FindDirectoriesToExclude(localpythonpathdir):
        logging.info(
            "Looking in '{0}' for directories to skip".format(localpythonpathdir)
        )
        xd_string = " /XD $TF /XD .git"
        for root, dir, files in os.walk(localpythonpathdir):
            for file in files:
                if file.lower() == ".ignoretgzchange":
                    xd_string += " /XD {0}".format(root)
        return xd_string

    def CopySource(self, localpythonpath, run_dir_abs):
        if self.update_remote_python_parent:
            remote_python_parent = self.remote_python_parent
        else:
            remote_python_parent = run_dir_abs + os.path.sep + "pythonpath"
        pstutil.create_directory_if_necessary(remote_python_parent, isfile=False)
        remotepythonpath_list = []
        for i, localpythonpathdir in enumerate(localpythonpath.split(";")):
            remotepythonpathdir = os.path.join(remote_python_parent, str(i))
            remotepythonpath_list.append(remotepythonpathdir)
            xd_string = HPC.FindDirectoriesToExclude(localpythonpathdir)
            xcopycommand = "robocopy /s {0} {1}{2}".format(
                localpythonpathdir, remotepythonpathdir, xd_string
            )
            logging.info(xcopycommand)
            os.system(xcopycommand)

        remotepythonpath = ";".join(remotepythonpath_list)
        return remotepythonpath

    def create_bat_file(
        self,
        distributable,
        remotepythoninstall,
        remotepythonpath,
        remotewd,
        run_dir_abs,
        run_dir_rel,
        result_remote,
        nodelocalwd,
        create_bat_file,
    ):
        path_share_list = [r"", r"Scripts"]
        remotepath_list = []
        for path_share in path_share_list:
            path_share_abs = os.path.join(remotepythoninstall, path_share)
            if not os.path.isdir(path_share_abs):
                raise Exception("Expect path directory at '{0}'".format(path_share_abs))
            remotepath_list.append(path_share_abs)
        remotepath = ";".join(remotepath_list)

        (
            distributablep_filename_rel,
            distributablep_filename_abs,
        ) = self.create_distributablep(distributable, run_dir_abs, run_dir_rel)

        distributable_py_file = os.path.join(
            os.path.dirname(__file__), "..", "distributable.py"
        )
        if not os.path.exists(distributable_py_file):
            raise Exception(
                "Expect file at " + distributable_py_file + ", but it doesn't exist."
            )
        localfilepath, file = os.path.split(distributable_py_file)

        for remote_path_part in remotepythonpath.split(";"):
            remoteexe = os.path.join(remote_path_part, "fastlmm", "util", file)
            if os.path.exists(remoteexe):
                break  # not continue
            remoteexe = None
        assert (
            remoteexe is not None
        ), "Could not find '{0}' on remote python path. Is fastlmm on your local python path?".format(
            file
        )

        # run_dir_rel + os.path.sep + "pythonpath" + os.path.sep + os.path.splitdrive(localfilepath)[1]

        # result_remote2 = result_remote.encode("string-escape")
        command_string = (
            remoteexe
            + r""" "{0}" """.format(distributablep_filename_abs)
            + r""" "LocalInParts(%1,{0},mkl_num_threads={1},result_file=""{2}"",run_dir=""{3}"") " """.format(
                self.taskcount,
                self.mkl_num_threads,
                "result.p",
                run_dir_abs.encode("string-escape"),
            )
        )
        batfilename_rel = os.path.join(run_dir_rel, "dist.bat")
        batfilename_abs = os.path.join(run_dir_abs, "dist.bat")
        pstutil.create_directory_if_necessary(batfilename_abs, isfile=True)
        matplotlibfilename_rel = os.path.join(run_dir_rel, ".matplotlib")
        matplotlibfilename_abs = os.path.join(run_dir_abs, ".matplotlib")
        pstutil.create_directory_if_necessary(matplotlibfilename_abs, isfile=False)
        pstutil.create_directory_if_necessary(
            matplotlibfilename_abs + "/tex.cache", isfile=False
        )
        ipythondir_rel = os.path.join(run_dir_rel, ".ipython")
        ipythondir_abs = os.path.join(run_dir_abs, ".ipython")
        pstutil.create_directory_if_necessary(ipythondir_abs, isfile=False)
        with open(batfilename_abs, "w") as batfile:
            batfile.write("set path={0};%path%\n".format(remotepath))
            batfile.write("set PYTHONPATH={0}\n".format(remotepythonpath))
            batfile.write("set USERPROFILE={0}\n".format(run_dir_abs))
            batfile.write("set MPLCONFIGDIR={0}\n".format(matplotlibfilename_abs))
            batfile.write("set IPYTHONDIR={0}\n".format(ipythondir_abs))
            batfile.write("python {0}\n".format(command_string))

        if self.node_local:
            with open(os.path.join(run_dir_abs, "nodeprep.bat"), "w") as prepfile:
                prepfile.write(r"""set f="{0}"{1}""".format(remotewd, "\n"))
                prepfile.write(r"""set t="{0}"{1}""".format(nodelocalwd, "\n"))
                prepfile.write("if not exist %t% mkdir %t%\n")
                with open(
                    os.path.join(run_dir_abs, "noderelease.bat"), "w"
                ) as releasefile:
                    releasefile.write(r"""set f="{0}"{1}""".format(remotewd, "\n"))
                    releasefile.write(r"""set t="{0}"{1}""".format(nodelocalwd, "\n"))
                    inputOutputCopier = HPCCopierNodeLocal(
                        prepfile, releasefile, self.clean_up
                    )  # Create the object that copies input and output files to where they are needed
                    inputOutputCopier.input(
                        distributable
                    )  # copy of the input files to where they are needed (i.e. to the cluster)
                    inputOutputCopier.output(
                        distributable
                    )  # copy of the output files to where they are needed (i.e. off the cluster)
                    releasefile.write("rmdir /s %t%\n")
                    releasefile.write("exit /b 0\n")

        return batfilename_rel

    def check_remote_pythoninstall(self):
        remotepythoninstall = (
            r"\\GCR\Scratch\RR1\escience\pythonInstallD"  # !!! don't hardwire this
        )
        if not os.path.isdir(remotepythoninstall):
            raise Exception(
                "Expect Python and related directories at '{0}'".format(
                    remotepythoninstall
                )
            )

        return remotepythoninstall

    def create_run_dir(self):
        username = os.environ["USERNAME"]
        localwd = os.getcwd()
        # !!make an option to specify the full remote WD. Also what is the "\\\\" case for?
        if localwd.startswith("\\\\"):
            remotewd = (
                self.fileshare
                + os.path.sep
                + username
                + os.path.sep
                + "\\".join(localwd.split("\\")[4:])
            )
            nodelocalwd = (
                "d:\scratch\escience"
                + os.path.sep
                + username
                + os.path.sep
                + "\\".join(localwd.split("\\")[4:])
            )  # !!!const
        else:
            remotewd = (
                self.fileshare + os.path.sep + username + os.path.splitdrive(localwd)[1]
            )  # using '+' because 'os.path.join' isn't work with shares
            nodelocalwd = (
                "d:\scratch\escience"
                + os.path.sep
                + username
                + os.path.splitdrive(localwd)[1]
            )  # !!! const
        import datetime

        now = datetime.datetime.now()
        run_dir_rel = os.path.join("runs", pstutil._datestamp(appendrandom=True))
        run_dir_abs = os.path.join(remotewd, run_dir_rel)
        pstutil.create_directory_if_necessary(run_dir_abs, isfile=False)

        return remotewd, run_dir_abs, run_dir_rel, nodelocalwd


class HPCCopier(object):  # Implements ICopier
    def __init__(self, remotewd, skipinput=False):
        self.remotewd = remotewd
        self.skipinput = skipinput

    def input(self, item):
        if self.skipinput:
            return
        if isinstance(item, str):
            itemnorm = os.path.normpath(item)
            remote_file_name = os.path.join(self.remotewd, itemnorm)
            remote_dir_name, ignore = os.path.split(remote_file_name)
            pstutil.create_directory_if_necessary(remote_file_name)
            xcopycommand = "xcopy /d /e /s /c /h /y {0} {1}".format(
                itemnorm, remote_dir_name
            )
            logging.info(xcopycommand)
            rc = os.system(xcopycommand)
            print("rc=" + str(rc))
            if rc != 0:
                raise Exception(
                    "xcopy cmd failed with return value={0}, from cmd {1}".format(
                        rc, xcopycommand
                    )
                )
        elif hasattr(item, "copyinputs"):
            item.copyinputs(self)
        # else -- do nothing

    def output(self, item):
        if isinstance(item, str):
            itemnorm = os.path.normpath(item)
            pstutil.create_directory_if_necessary(itemnorm)
            remote_file_name = os.path.join(self.remotewd, itemnorm)
            local_dir_name, ignore = os.path.split(itemnorm)
            assert os.path.exists(
                remote_file_name
            ), "Don't see expected file '{0}'. Did the HPC job fail?".format(
                remote_file_name
            )
            # xcopycommand = "xcopy /d /e /s /c /h /y {0} {1}".format(remote_file_name, local_dir_name) # we copy to the local dir instead of the local file so that xcopy won't ask 'file or dir?'
            xcopycommand = "xcopy /d /c /y {0} {1}".format(
                remote_file_name, local_dir_name
            )  # we copy to the local
            logging.info(xcopycommand)
            rc = os.system(xcopycommand)
            if rc != 0:
                logging.info(
                    "xcopy cmd failed with return value={0}, from cmd {1}".format(
                        rc, xcopycommand
                    )
                )
        elif hasattr(item, "copyoutputs"):
            item.copyoutputs(self)
        # else -- do nothing


class HPCCopierNodeLocal(object):  # Implements ICopier
    def __init__(self, fileprep, filerelease, clean_up):
        self.fileprep = fileprep
        self.filerelease = filerelease
        self.clean_up = clean_up

    def input(self, item):
        if isinstance(item, str):
            itemnorm = os.path.normpath(item)
            dirname = os.path.dirname(itemnorm)
            self.fileprep.write("if not exist %t%\{0} mkdir %t%\{0}\n".format(dirname))
            self.fileprep.write(
                "xcopy /d /e /s /c /h /y %f%\{0} %t%\{1}\\\n".format(itemnorm, dirname)
            )
            if self.clean_up:
                self.filerelease.write("del %t%\{0}\n".format(itemnorm))
        elif hasattr(item, "copyinputs"):
            item.copyinputs(self)
        # else -- do nothing

    def output(self, item):
        if isinstance(item, str):
            itemnorm = os.path.normpath(item)
            dirname = os.path.dirname(itemnorm)
            self.filerelease.write(
                "xcopy /d /e /s /c /h /y %t%\{0} %f%\{1}\\\n".format(itemnorm, dirname)
            )
            if self.clean_up:
                self.filerelease.write("del %t%\{0}\n".format(itemnorm))
        elif hasattr(item, "copyoutputs"):
            item.copyoutputs(self)
        # else -- do nothing



import numpy as np
import subprocess, sys, os.path
from itertools import *
import pandas as pd
import logging
from pysnptools.snpreader import SnpReader
from pysnptools.snpreader import SnpData
import warnings
from pysnptools.pstreader import _OneShot

#LATER say that this is also called dosage or bimbam
#LATER work when the extension is \*.dosage or anything else
#LATER make so fam can be called bam and map can be called bim
class Dat(_OneShot,SnpReader):
    '''
    A :class:`.SnpReader` for reading Dat/Fam/Map-formated files from disk.

    See :class:`.SnpReader` for general examples of using SnpReaders.

    This is a text format that can store any numeric values. (In contrast, Bed and Ped can only store 0,1,2, and missing). Its Dat files look like::
    
        null_0  	j	n	0.333	1	2
        null_100	j	n	2	1	1
        null_200	j	n	0	nan	1
        ...

    Its Map and Fam files are described `here <http://zzz.bwh.harvard.edu/plink/data.shtml>`__.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The Dat file to read.
                     * **skiprows** (*int*) -- Number of lines to skip before reading. Defaults to 0.

        :Example:

        >>> from pysnptools.snpreader import Dat
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> dat_file = example_file("pysnptools/examples/toydata.*","*.dat")
        >>> data_on_disk = Dat(dat_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 10000)

    **Methods beyond** :class:`.SnpReader`
    '''

    def __init__(self, filename, skiprows=0):
        '''
        filename    : string of the name of the Dat file.
        '''
        super(Dat, self).__init__()
        self.filename = SnpReader._name_of_other_file(filename,remove_suffix="dat", add_suffix="dat")
        self.skiprows = skiprows

    def _read_pstdata(self):
        row = SnpReader._read_fam(self.filename,remove_suffix="dat")
        col, col_property = SnpReader._read_map_or_bim(self.filename,remove_suffix="dat", add_suffix="map")
        if len(row)==0 or len(col)==0:
            return SnpData(iid=row,sid=col,pos=col_property,val=np.empty([len(row),len(col)]))
        datfields = pd.read_csv(self.filename,delimiter = '\t',header=None,index_col=False,skiprows=self.skiprows)
        if not np.array_equal(datfields[0], col) : raise Exception("Expect snp list in map file to exactly match snp list in dat file")
        del datfields[0]
        del datfields[1]
        del datfields[2]
        assert len(row) == datfields.shape[1], "Expect # iids in fam file to match dat file"
        val = datfields.values.T
        snpdata = SnpData(iid=row,sid=col,pos=col_property,val=val)
        return snpdata

    def copyinputs(self, copier):
        # doesn't need to self.run_once() because creates name of all files itself
        copier.input(SnpReader._name_of_other_file(self.filename,remove_suffix="dat", add_suffix="dat"))
        copier.input(SnpReader._name_of_other_file(self.filename,remove_suffix="dat", add_suffix="fam"))
        copier.input(SnpReader._name_of_other_file(self.filename,remove_suffix="dat", add_suffix="map"))

    @staticmethod
    def write(filename, snpdata):
        """Writes a :class:`SnpData` to dat/fam/map format and returns the :class:`.Dat`.

        :param filename: the name of the file to create
        :type filename: string
        :param snpdata: The in-memory data that should be written to disk.
        :type snpdata: :class:`SnpData`
        :rtype: :class:`.Dat`

        >>> from pysnptools.snpreader import Dat, Bed
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
        >>> snpdata = Bed(bed_file,count_A1=False)[:,:10].read()  # Read first 10 snps from Bed format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.dat")
        >>> Dat.write("tempdir/toydata10.dat",snpdata)              # Write data in dat/fam/map format
        Dat('tempdir/toydata10.dat')
        """

        if isinstance(filename,SnpData) and isinstance(snpdata,str): #For backwards compatibility, reverse inputs if necessary
            warnings.warn("write statement should have filename before data to write", DeprecationWarning)
            filename, snpdata = snpdata, filename 

        SnpReader._write_fam(snpdata, filename, remove_suffix="dat")
        SnpReader._write_map_or_bim(snpdata, filename, remove_suffix="dat", add_suffix="map")
        filename = SnpReader._name_of_other_file(filename,remove_suffix="dat", add_suffix="dat")

        snpsarray = snpdata.val
        with open(filename,"w") as dat_filepointer:
            for sid_index, sid in enumerate(snpdata.sid):
                if sid_index % 1000 == 0:
                    logging.info("Writing snp # {0} to file '{1}'".format(sid_index, filename))
                dat_filepointer.write("%s\tj\tn\t"%sid) #use "j" and "n" as the major and minor allele
                row = snpsarray[:,sid_index]
                dat_filepointer.write("\t".join((str(i) for i in row)) + "\n")
        logging.info("Done writing " + filename)
        return Dat(filename)

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""Tools for reading and manipulating kernels. A kernels is a matrix from iid to iid that typically tells the relatedness of pairs of people.
"""

from pysnptools.kernelreader.kernelreader import KernelReader
from pysnptools.kernelreader.kerneldata import KernelData
from pysnptools.kernelreader.snpkernel import SnpKernel
from pysnptools.kernelreader.identity import Identity
from pysnptools.kernelreader.kernelnpz import KernelNpz
from pysnptools.kernelreader.kernelhdf5 import KernelHdf5


import numpy as np
import os
import os.path
from itertools import *
import pandas as pd
import logging
import time
import pysnptools.util as pstutil
from pysnptools.pstreader import PstReader
from pysnptools.snpreader import SnpData
import warnings
import pysnptools.standardizer as stdizer
from pysnptools.snpreader._dist2snp import _Dist2Snp


# !!why do the examples use ../tests/datasets instead of "examples"?
class DistReader(PstReader):
    """A DistReader is one of three things:

    * A class such as :class:`.Bgen` for you to specify a file with data. For example,

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> dist_on_disk = Bgen(bgen_file)
        >>> print(dist_on_disk) # prints the name of the file reader
        Bgen('...pysnptools/examples/2500x100.bgen')
        >>> dist_on_disk.sid_count # prints the number of SNPS (but doesn't read any SNP distribution values)
        100

    * A :class:`.DistData` class that holds SNP distribution data in memory, typically after reading it from disk:

        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> dist_on_disk = Bgen(bgen_file)
        >>> distdata1 = dist_on_disk.read() #reads the SNP distribution values
        >>> print(type(distdata1.val)) # The val property is an 3-D ndarray of SNP distribution values
        <class 'numpy.ndarray'>
        >>> print(distdata1) # prints the name in-memory SNP distribution reader.
        DistData(Bgen('...pysnptools/examples/2500x100.bgen'))
        >>> distdata1.iid_count #prints the number of iids (number of individuals) in this in-memory data
        2500

    * A subset of any DistReader, specified with "[ *iid_index* , *sid_index* ]", to read only some SNP distribution values. It can
      also be used to re-order the values.

        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> dist_on_disk = Bgen(bgen_file)
        >>> subset_on_disk = dist_on_disk[[3,4],::2] # specification for a subset of the data on disk. No SNP distriubtion values are read yet.
        >>> print(subset_on_disk.sid_count) # prints the number of sids in this subset (but still doesn't read any SNP distribution values)
        50
        >>> print(subset_on_disk) #prints a specification of 'subset_on_disk'
        Bgen('...pysnptools/examples/2500x100.bgen')[[3,4],::2]
        >>> distdata_subset = subset_on_disk.read() # efficiently reads the specified subset of values from the disk
        >>> print(distdata_subset) # prints the specification of the in-memory SNP distribution information
        DistData(Bgen('...pysnptools/examples/2500x100.bgen')[[3,4],::2])
        >>> print((int(distdata_subset.val.shape[0]), int(distdata_subset.val.shape[1]))) # The dimensions of the ndarray of SNP distriubtion values
        (2, 50)

    The DistReaders Classes

        ========================= =================== ====================== ================== ======================
        *Class*                   *Format*            *Random Access*        *Suffixes*         *Write* method?
        :class:`.DistData`        in-memory floats    Yes                    *n/a*              *n/a*
        :class:`.Bgen`            binary, floats      Yes (by sid)           \*.bgen            Yes
        :class:`.DistNpz`         binary, floats      No                     .dist.npz          Yes
        :class:`.DistHdf5`        binary, floats      Yes (by sid or iid)    .dist.hdf5         Yes
        :class:`.DistMemMap`      mem-mapped floats   Yes                    .dist.memmap       Yes
        ========================= =================== ====================== ================== ======================


    Methods & Properties:

        Every DistReader, such as :class:`.Bgen` and :class:`.DistData`, has these properties: :attr:`iid`, :attr:`iid_count`, :attr:`sid`, :attr:`sid_count`,
        :attr:`pos` and these methods: :meth:`read`, :meth:`iid_to_index`, :meth:`sid_to_index`, :meth:`as_snp`. See below for details.

        :class:`.DistData` is a DistReader so it supports the above properties and methods. In addition, it supports property :attr:`DistData.val`.

        See below for details.

        Many of the classes, such as :class:`.Bgen`, also provide a static :meth:`Bgen.write` method for writing :class:`.DistReader` or :class:`.DistData` to disk.

        >>> from pysnptools.distreader import DistHdf5, Bgen
        >>> import pysnptools.util as pstutil
        >>> hdf5_file = example_file("pysnptools/examples/toydata.snpmajor.dist.hdf5")
        >>> distreader = DistHdf5(hdf5_file)[:,:10] # A reader for the first 10 SNPs in Hdf5 format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.bgen")
        >>> Bgen.write("tempdir/toydata10.bgen",distreader)        # Write data in BGEN format
        Bgen('tempdir/toydata10.bgen')


    iids and sids:

        *same as with* :class:`.SnpReader`

    Selecting and Reordering Individuals and SNPs

        *same as with* :class:`.SnpReader`

    When Data is Read:

        *same as with* :class:`.SnpReader`

    When Data is Re-Read and Copied:

        *same as with* :class:`.SnpReader`

    Avoiding Unwanted ndarray Allocations

        *same as with* :class:`.SnpReader`

    The :meth:`read` Method

        By default the :meth:`read` returns a 3-D ndarray of numpy.float64 laid out in memory in F-contiguous order
        (iid-index varies the fastest). You may, instead,
        ask for numpy.float32 or for C-contiguous order or any order. See :meth:`read` for details.


    Details of Methods & Properties:
    """

    def __init__(self, *args, **kwargs):
        super(DistReader, self).__init__(*args, **kwargs)

    @property
    def iid(self):
        """A ndarray of the iids. Each iid is a ndarray of two strings (a family ID and a individual ID) that identifies an individual.

        :rtype: ndarray of strings with shape [:attr:`.iid_count`,2]

        This property (to the degree practical) reads only iid and sid data from the disk, not SNP distribution data. Moreover, the iid and sid data is read from file only once.

        :Example:

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> dist_on_disk = Bgen(bgen_file)
        >>> print(dist_on_disk.iid[:3]) # print the first three iids
        [['0' 'iid_0']
         ['0' 'iid_1']
         ['0' 'iid_2']]
        """
        return self.row

    @property
    def iid_count(self):
        """number of iids

        :rtype: integer

        This property (to the degree practical) reads only iid and sid data from the disk, not SNP distribution data. Moreover, the iid and sid data is read from file only once.
        """
        return self.row_count

    @property
    def sid(self):
        """A ndarray of the sids. Each sid is a string that identifies a SNP.

        :rtype: ndarray (length :attr:`.sid_count`) of strings

        This property (to the degree practical) reads only iid and sid data from the disk, not SNP distribution data. Moreover, the iid and sid data is read from file only once.

        :Example:

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> dist_on_disk = Bgen(bgen_file)
        >>> print(dist_on_disk.sid[:9]) # print the first nine sids
        ['sid_0' 'sid_1' 'sid_2' 'sid_3' 'sid_4' 'sid_5' 'sid_6' 'sid_7' 'sid_8']
        """
        return self.col

    @property
    def sid_count(self):
        """number of sids

        :rtype: integer

        This property (to the degree practical) reads only iid and sid data from the disk, not SNP distribution data. Moreover, the iid and sid data is read from file only once.

        """
        return self.col_count

    # !!document that chr must not be X,Y,M only numbers (as per the PLINK DistNpz format)
    # !!Also what about telling the ref and alt allele? Also, what about tri and quad alleles, etc?
    @property
    def pos(self):
        """A ndarray of the position information for each sid. Each element is a ndarray of three numpy.numbers (chromosome, genetic distance, basepair distance).

        :rtype: ndarray of float64 with shape [:attr:`.sid_count`, 3]

        This property (to the degree practical) reads only iid and sid data from the disk, not SNP distribution data. Moreover, the iid and sid data is read from file only once.

        :Example:

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> dist_on_disk = Bgen(bgen_file)
        >>> print(dist_on_disk.pos[:4,].astype('int')) # print position information for the first four sids: #The '...' is for possible space char
        [[        1         0   9270273]
         [        1         0  39900273]
         [        1         0  70530273]
         [        1         0 101160273]]
        """
        return self.col_property

    @property
    def row_property(self):
        """Defined as a zero-width array for compatibility with :class:`PstReader`, but not used."""
        if not hasattr(self, "_row_property"):
            self._row_property = np.empty((self.row_count, 0))
        return self._row_property

    def _read(
        self,
        iid_index_or_none,
        sid_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        raise NotImplementedError

    # !!check that views always return contiguous memory by default
    def read(
        self,
        order="F",
        dtype=np.float64,
        force_python_only=False,
        view_ok=False,
        num_threads=None,
    ):
        """Reads the SNP values and returns a :class:`.DistData` (with :attr:`DistData.val` property containing a new 3D ndarray of the SNP distribution values).

        :param order: {'F' (default), 'C', 'A'}, optional -- Specify the order of the ndarray. If order is 'F' (default),
            then the array will be in F-contiguous order (iid-index varies the fastest).
            If order is 'C', then the returned array will be in C-contiguous order (sid-index varies the fastest).
            If order is 'A', then the :attr:`DistData.val`
            ndarray may be in any order (either C-, Fortran-contiguous).
        :type order: string or None

        :param dtype: {numpy.float64 (default), numpy.float32}, optional -- The data-type for the :attr:`DistData.val` ndarray.
        :type dtype: data-type

        :param force_python_only: optional -- If False (default), may use outside library code. If True, requests that the read
            be done without outside library code.
        :type force_python_only: bool

        :param view_ok: optional -- If False (default), allocates new memory for the :attr:`DistData.val`'s ndarray. If True,
            if practical and reading from a :class:`DistData`, will return a new
            :class:`DistData` with a ndarray shares memory with the original :class:`DistData`.
            Typically, you'll also wish to use "order='A'" to increase the chance that sharing will be possible.
            Use these parameters with care because any change to either ndarraywill effect
            the others. Also keep in mind that :meth:`read` relies on ndarray's mechanisms to decide whether to actually
            share memory and so it may ignore your suggestion and allocate a new ndarray anyway.
        :type view_ok: bool

        :param num_threads: optional -- The number of threads with which to read data. Defaults to all available
            processors. Can also be set with these environment variables (listed in priority order):
            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
        :type num_threads: None or int

        :rtype: :class:`.DistData`

        Calling the method again causes the SNP distribution values to be re-read and creates a new in-memory :class:`.DistData` with a new ndarray of SNP values.

        If you request the values for only a subset of the sids or iids, (to the degree practical) only that subset will be read from disk.

        :Example:

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> dist_on_disk = Bgen(bgen_file) # Specify SNP data on disk
        >>> distdata1 = dist_on_disk.read() # Read all the SNP data returning a DistData instance
        >>> print(type(distdata1.val).__name__) # The DistData instance contains a ndarray of the data.
        ndarray
        >>> subset_distdata = dist_on_disk[:,::2].read() # From the disk, read SNP values for every other sid
        >>> print(subset_distdata.val[0,0]) # Print the first SNP value in the subset
        [0.466804   0.38812848 0.14506752]
        >>> subsub_distdata = subset_distdata[:10,:].read(order='A',view_ok=True) # Create an in-memory subset of the subset with SNP values for the first ten iids. Share memory if practical.
        >>> import numpy as np
        >>> # print np.may_share_memory(subset_distdata.val, subsub_distdata.val) # Do the two ndarray's share memory? They could. Currently they won't.
        """
        dtype = np.dtype(dtype)
        val = self._read(
            None, None, order, dtype, force_python_only, view_ok, num_threads
        )
        from pysnptools.distreader import DistData

        ret = DistData(self.iid, self.sid, val, pos=self.pos, name=str(self))
        return ret

    def as_snp(self, max_weight=2.0, block_size=None):
        """Returns a :class:`pysnptools.snpreader.SnpReader` such that turns the probability distribution into an expected value.

        For example, if the probability distribution is [0.466804   0.38812848 0.14506752] and the max_weight is 2, then the expected
        value will be (0.466804*0+0.38812848*.5+0.14506752*1)*2, 0.67826352 .

        :param max_weight: optional -- The expected values will range from 0 to *max_weight*. Default of 2.
        :type max_weight: number

        :param block_size: optional -- Default of None (meaning to load all). Suggested number of sids to read into memory at a time.
        :type block_size: int or None

        :rtype: class:`SnpReader`

        :Example:

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> distreader = Bgen(bgen_file) # Specify distribution data on disk
        >>> print(distreader[0,0].read().val)
        [[[0.466804   0.38812848 0.14506752]]]
        >>> snpreader = distreader.as_snp(max_weight=2)
        >>> print(snpreader[0,0].read().val)
        [[0.67826352]]
        """
        dist2snp = _Dist2Snp(self, max_weight=max_weight, block_size=block_size)
        return dist2snp

    def iid_to_index(self, list):
        """Takes a list of iids and returns a list of index numbers

        :param list: list of iids
        :type order: list of list of strings

        :rtype: ndarray of int

        This method (to the degree practical) reads only iid and sid data from the disk, not SNP value data. Moreover, the iid and sid data is read from file only once.

        :Example:

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> dist_on_disk = Bgen(bgen_file) # Specify SNP data on disk
        >>> print(dist_on_disk.iid_to_index([['0','iid_2'],['0','iid_1']])) #Find the indexes for two iids.
        [2 1]
        """
        return self.row_to_index(list)

    def sid_to_index(self, list):
        """Takes a list of sids and returns a list of index numbers

        :param list: list of sids
        :type list: list of strings

        :rtype: ndarray of int

        This method (to the degree practical) reads only iid and sid data from the disk, not SNP value data. Moreover, the iid and sid data is read from file only once.

        :Example:

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> dist_on_disk = Bgen(bgen_file) # Specify SNP data on disk
        >>> print(dist_on_disk.sid_to_index(['sid_2','sid_9'])) #Find the indexes for two sids.
        [2 9]
        """
        return self.col_to_index(list)

    @property
    def val_shape(self):
        """
        Tells the shape of value for a given individual and SNP. For DistReaders always returns 3.
        """
        return 3

    def __getitem__(self, iid_indexer_and_snp_indexer):
        from pysnptools.distreader._subset import _DistSubset

        iid_indexer, snp_indexer = iid_indexer_and_snp_indexer
        return _DistSubset(self, iid_indexer, snp_indexer)

    @staticmethod
    def _as_distdata(distreader, force_python_only, order, dtype, num_threads):
        """
        Like 'read' except won't read if already a DistData
        """
        from pysnptools.distreader import DistData  # must import here to avoid cycles

        dtype = np.dtype(dtype)

        if (
            hasattr(distreader, "val")
            and distreader.val.dtype == dtype
            and (
                order == "A"
                or (order == "C" and distreader.val.flags["C_CONTIGUOUS"])
                or (order == "F" and distreader.val.flags["F_CONTIGUOUS"])
            )
        ):
            return distreader
        else:
            return distreader.read(
                order=order, dtype=dtype, view_ok=True, num_threads=num_threads
            )

    def copyinputs(self, copier):
        raise NotImplementedError

    def _assert_iid_sid_pos(self, check_val):
        if check_val:
            assert (
                len(self._val.shape) == 3 and self._val.shape[-1] == 3
            ), "val should have 3 dimensions and the last dimension should have size 3"
            assert self._val.shape == (
                len(self._row),
                len(self._col),
                3,
            ), "val shape should match that of iid_count x sid_count"
        assert (
            self._row.dtype.type is np.str_
            and len(self._row.shape) == 2
            and self._row.shape[1] == 2
        ), "iid should be dtype str, have two dimensions, and the second dimension should be size 2"
        assert (
            self._col.dtype.type is np.str_ and len(self._col.shape) == 1
        ), "sid should be of dtype of str and one dimensional"


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    if False:
        from pysnptools.distreader import Bgen

        dist_on_disk = Bgen("../examples/2500x100.bgen")
        print(
            dist_on_disk.pos[:4,].astype("int")
        )  # print position information for the first three sids: #The '...' is for possible space char

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc t
    print("done")



from pysnptools.util.mapreduce1.runner import Runner, _JustCheckExists, _run_one_task
import os
import logging
import cloudpickle as pickle
import subprocess, sys, os.path
import multiprocessing
from pysnptools.util import create_directory_if_necessary, _datestamp


class LocalMultiProc(Runner):
    """
    A :class:`.Runner` that runs a :func:`.map_reduce` as multiple processes on a single machine.

    **Constructor:**
        :Parameters: * **taskcount** (*number*) -- The number of processes to run on.
                     * **mkl_num_threads** (*number*) -- (default None) Limit on the number threads used by the NumPy MKL library.
                     * **weights** (*array of integers*) -- (default None) If given, tells the relative amount of work assigned to
                            each task. The length of the array must be **taskcount**. If not given, all tasks are assigned the
                            same amount of work.
                     * **taskindex_to_environ** (*function from integers to dictionaries*) -- (default None). If given, this
                            should be function from taskindex to a dictionary. The dictionary tells how to temporarily set
                            environment variables while the task is running. The dictionary is a mapping of
                            variables and values (both strings).
                     * **just_one_process** (*bool*) -- (default False) Divide the work for multiple processes, but runs sequentially on one process. Can be useful for debugging.
                     * **logging_handler** (*stream*) --  (default stdout) Where to output logging messages.

        :Example:

        >>> from pysnptools.util.mapreduce1 import map_reduce
        >>> from pysnptools.util.mapreduce1.runner import LocalMultiProc
        >>> def holder1(n,runner):
        ...     def mapper1(x):
        ...         return x*x
        ...     def reducer1(sequence):
        ...        return sum(sequence)
        ...     return map_reduce(range(n),mapper=mapper1,reducer=reducer1,runner=runner)
        >>> holder1(100,LocalMultiProc(4))
        328350

    """

    def __init__(
        self,
        taskcount,
        mkl_num_threads=None,
        weights=None,
        taskindex_to_environ=None,
        just_one_process=False,
        logging_handler=logging.StreamHandler(sys.stdout),
    ):
        self.just_one_process = just_one_process
        logger = logging.getLogger()
        if not logger.handlers:
            logger.setLevel(logging.INFO)
        for h in list(logger.handlers):
            logger.removeHandler(h)
        if logger.level == logging.NOTSET:
            logger.setLevel(logging.INFO)
        logger.addHandler(logging_handler)

        self.taskcount = taskcount
        self.mkl_num_threads = mkl_num_threads
        self.weights = weights
        self.taskindex_to_environ = taskindex_to_environ

    def run(self, distributable):
        _JustCheckExists().input(distributable)

        localpath = os.environ["PATH"]
        localwd = os.getcwd()

        import datetime

        now = datetime.datetime.now()
        run_dir_rel = os.path.join("runs", _datestamp(appendrandom=True))
        run_dir_abs = os.path.join(localwd, run_dir_rel)
        create_directory_if_necessary(run_dir_rel, isfile=False)

        distributablep_filename = os.path.join(run_dir_rel, "distributable.p")
        with open(distributablep_filename, mode="wb") as f:
            pickle.dump(distributable, f)

        distributable_py_file = os.path.join(
            os.path.dirname(__file__), "..", "distributable.py"
        )
        if not os.path.exists(distributable_py_file):
            raise Exception(
                "Expect file at " + distributable_py_file + ", but it doesn't exist."
            )

        if not self.just_one_process:
            proc_list = []
            for taskindex in range(self.taskcount):
                environ = (
                    self.taskindex_to_environ(taskindex)
                    if self.taskindex_to_environ is not None
                    else None
                )
                command_string_list = [
                    sys.executable,
                    distributable_py_file,
                    distributablep_filename,
                    f"LocalInParts({taskindex},{self.taskcount},mkl_num_threads={self.mkl_num_threads},weights={self.weights},environ={environ})".replace(
                        " ", ""
                    ),
                ]

                # logging.info(command_string_list)
                proc = subprocess.Popen(command_string_list, cwd=os.getcwd())
                proc_list.append(proc)

            for taskindex, proc in enumerate(proc_list):
                rc = proc.wait()
                if not 0 == rc:
                    raise Exception(
                        "Running python in python results in non-zero return code in task#{0}".format(
                            taskindex
                        )
                    )
        else:
            from pysnptools.util.mapreduce1.runner import LocalInParts

            for taskindex in range(self.taskcount):
                environ = (
                    self.taskindex_to_environ(taskindex)
                    if self.taskindex_to_environ is not None
                    else None
                )
                LocalInParts(
                    taskindex,
                    self.taskcount,
                    mkl_num_threads=self.mkl_num_threads,
                    weights=self.weights,
                    environ=environ,
                ).run(distributable)

        environ = (
            self.taskindex_to_environ(self.taskcount)
            if self.taskindex_to_environ is not None
            else None
        )
        result = _run_one_task(
            distributable,
            self.taskcount,
            self.taskcount,
            distributable.tempdirectory,
            weights=self.weights,
            environ=environ,
        )

        _JustCheckExists().output(distributable)
        return result


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    from pysnptools.util.mapreduce1 import (
        map_reduce,
    )  # Needed to work around thread local variable issue

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
import logging
from pysnptools.standardizer import Standardizer

class Identity(Standardizer):
    '''
    A :class:`.Standardizer` that does nothing to SNP data.

    See :class:`.Standardizer` for more information about standardization.

    >>> from pysnptools.standardizer import Identity
    >>> from pysnptools.snpreader import Bed
    >>> from pysnptools.util import example_file # Download and return local file name
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> snpdata1 = Bed(bedfile,count_A1=False).read()
    >>> print(snpdata1.val[0,0])
    2.0
    >>> snpdata1 = snpdata1.standardize(Identity())
    >>> print(snpdata1.val[0,0])
    2.0
    '''

    def __init__(self):
        super(Identity, self).__init__()

    def standardize(self, snps, block_size=None, return_trained=False, force_python_only=False, num_threads=None):
        if block_size is not None:
            warnings.warn("block_size is deprecated (and not needed, since standardization is in-place", DeprecationWarning)
        if return_trained:
            return snps, self
        else:
            return snps

    @property
    def is_constant(self):
        return True        

    def __repr__(self): 
        return "{0}()".format(self.__class__.__name__)

    def _merge_trained(self, trained_list):
        return self

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
import logging
from pysnptools.standardizer import Standardizer
import warnings

class BySidCount(Standardizer):
    '''
    '''

    def __init__(self):
        super(BySidCount, self).__init__()
        warnings.warn("BySidCount no longer supported", DeprecationWarning)

